# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "20/01/2020"


import unittest

import numpy

from silx.opencl.common import ocl

from darfix.core.componentsMatching import ComponentsMatching, Method
from skimage import data


class TestComponentsMatching(unittest.TestCase):

    """Tests for `componentsMatching.py`."""

    def setUp(self):
        self.components1 = numpy.array([data.moon(), data.camera(), data.gravel()])

        self.components2 = numpy.array(
            [self.components1[2], self.components1[0], self.components1[1]]
        )
        self.componentsMatching = ComponentsMatching(
            components=[self.components1, self.components2]
        )

    def test_euclidean_distance(self):
        ed = self.componentsMatching.euclidean_distance(
            self.components1[0], self.components2[1]
        )

        self.assertEqual(ed, 0)

        ed = self.componentsMatching.euclidean_distance(
            self.components1[1], self.components2[1]
        )

        self.assertNotEqual(ed, 0)

    @unittest.skipUnless(ocl, "PyOpenCl is missing")
    def test_sift_match(self):
        final_matches, matches = self.componentsMatching.match_components(
            method=Method.sift_feature_matching
        )
        self.assertEqual(final_matches[0], 1)

    def test_orb_match(self):
        final_matches, matches = self.componentsMatching.match_components(
            method=Method.orb_feature_matching
        )
        self.assertEqual(final_matches[0], 1)

    def test_draw_matches0(self):
        final_matches, matches = self.componentsMatching.match_components(
            method=Method.orb_feature_matching
        )
        stack = self.componentsMatching.draw_matches(
            final_matches, matches, displayMatches=True
        )
        self.assertEqual(stack[0].shape, (512, 1024))
        stack = self.componentsMatching.draw_matches(
            final_matches, matches, displayMatches=False
        )
        self.assertEqual(stack[1].shape, (512, 1024))

    def test_draw_matches1(self):
        final_matches, matches = self.componentsMatching.match_components(
            method=Method.euclidean_distance
        )
        stack = self.componentsMatching.draw_matches(final_matches, matches)
        self.assertEqual(stack[2].shape, (512, 1024))

    @unittest.skipUnless(ocl, "PyOpenCl is missing")
    def test_draw_matches2(self):
        final_matches, matches = self.componentsMatching.match_components(
            method=Method.sift_feature_matching
        )
        stack = self.componentsMatching.draw_matches(final_matches, matches)
        self.assertEqual(stack[2].shape, (512, 1024))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "11/05/2020"

import logging

import numpy

from silx.gui import qt
from silx.gui.plot import Plot1D

from .operationThread import OperationThread
from darfix.core.dataset import Operation

_logger = logging.getLogger(__file__)


class DataPartitionWidget(qt.QMainWindow):

    sigComputed = qt.Signal()

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        self._plot = Plot1D()

        binsLabel = qt.QLabel("Number of histogram bins:")
        filterRangeLabel = qt.QLabel("Filter range:")
        self.bins = qt.QLineEdit("")
        self.bins.setToolTip(
            "Defines the number of equal-width bins in the given range for the histogram"
        )
        self.bins.setValidator(qt.QIntValidator())
        self.bottomBinsNumber = qt.QLineEdit("")
        self.bottomBinsNumber.setPlaceholderText("First bin")
        self.bottomBinsNumber.setToolTip(
            "Minimum bin to use. It is 0 if nothing is entered."
        )
        self.topBinsNumber = qt.QLineEdit("")
        self.topBinsNumber.setPlaceholderText("Last bin")
        self.topBinsNumber.setToolTip(
            "Maximum bin to use. It is the number of bins if nothing is entered"
        )
        self.bottomBinsNumber.setValidator(qt.QIntValidator())
        self.topBinsNumber.setValidator(qt.QIntValidator())
        self.computeHistogram = qt.QPushButton("Compute histogram")
        self.computePartition = qt.QPushButton("Filter")
        self.abortB = qt.QPushButton("Abort")
        self.abortB.hide()
        self.computeHistogram.pressed.connect(self._computeHistogram)
        self.computePartition.pressed.connect(self._computePartition)
        self.abortB.pressed.connect(self.abort)
        widget = qt.QWidget(parent=self)
        layout = qt.QGridLayout()
        layout.addWidget(binsLabel, 0, 0, 1, 1)
        layout.addWidget(self.bins, 0, 1, 1, 1)
        layout.addWidget(filterRangeLabel, 1, 0, 1, 1)
        layout.addWidget(self.bottomBinsNumber, 1, 1)
        layout.addWidget(self.topBinsNumber, 1, 2)
        layout.addWidget(self.computeHistogram, 0, 2, 1, 1)
        layout.addWidget(self.computePartition, 2, 2, 1, 1)
        layout.addWidget(self.abortB, 3, 2, 1, 1)
        layout.addWidget(self._plot, 3, 0, 1, 3)
        widget.setLayout(layout)
        widget.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self.setCentralWidget(widget)
        self._plot.hide()

    def _computeHistogram(self):
        self.computeHistogram.setEnabled(False)
        self.computePartition.setEnabled(False)
        self.abortB.show()
        try:
            self._thread = OperationThread(self, self._dataset.compute_frames_intensity)
            self._thread.finished.connect(self._showHistogram)
            self._thread.start()
        except Exception as e:
            self.computeHistogram.setEnabled(True)
            raise e

    def _computePartition(self):
        self.computeHistogram.setEnabled(False)
        self.computePartition.setEnabled(False)
        self.abortB.show()
        try:
            self._thread = OperationThread(self, self._dataset.partition_by_intensity)
            self._thread.setArgs(
                bins=int(self.bins.text()),
                bottom_bin=int(self.bottomBinsNumber.text())
                if self.bottomBinsNumber.text() != ""
                else None,
                top_bin=int(self.topBinsNumber.text())
                if self.topBinsNumber.text() != ""
                else None,
            )
            self._thread.finished.connect(self._filterData)
            self._thread.start()
        except Exception as e:
            self.computePartition.setEnabled(True)
            raise e

    def abort(self):
        self.abortB.setEnabled(False)
        self._dataset.stop_operation(Operation.PARTITION)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter.

        :param Dataset dataset: dataset
        """
        self.parent = parent
        self._dataset = dataset
        self.indices = indices
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset
        self.bins.setText(str(self._dataset.nframes))

    def _updateDataset(self, widget, dataset):
        del self._dataset
        self._dataset = dataset
        self._update_dataset = dataset
        self.parent._updateDataset(widget, dataset)

    def _showHistogram(self):
        """
        Plots the eigenvalues.
        """
        self._thread.finished.disconnect(self._showHistogram)
        self.abortB.hide()
        self.abortB.setEnabled(True)
        self.computePartition.setEnabled(True)
        self.computeHistogram.setEnabled(True)
        if self._thread.data is not None:
            frames_intensity = self._thread.data
            self._plot.remove()
            self._plot.show()
            values, bins = numpy.histogram(frames_intensity, int(self.bins.text()))
            self._plot.addHistogram(values, bins, fill=True)
        else:
            print("\nComputation aborted")

    def _filterData(self):
        """
        Plots the eigenvalues.
        """
        self._thread.finished.disconnect(self._filterData)
        self.abortB.hide()
        self.abortB.setEnabled(True)
        self.computeHistogram.setEnabled(True)
        self.computePartition.setEnabled(True)
        if self._thread.data is not None:
            self.indices, self.bg_indices = self._thread.data
            self.sigComputed.emit()
        else:
            print("\nComputation aborted")

    def getDataset(self):
        return self._dataset, self.indices, self.bg_indices, self.bg_dataset



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "12/08/2019"


import copy

from Orange.widgets.widget import OWWidget, Input, Output
from darfix.core.process import DataCopy


class DataCopyWidgetOW(OWWidget):

    """
    Widget that creates a new dataset from a given one, and copies its data.
    """

    name = "data copy"
    icon = "icons/copy.svg"
    want_main_area = False
    ewokstaskclass = DataCopy

    # Inputs
    class Inputs:
        dataset = Input("dataset", tuple)

    # Outputs
    class Outputs:
        dataset = Output("dataset", tuple)

    def __init__(self):
        super().__init__()

    @Inputs.dataset
    def setDataset(self, _input):
        if _input is not None:
            # Copy and send new dataset
            dataset, update = _input
            if not update:
                self.cp_dataset = copy.deepcopy(dataset[1:])
                self.Outputs.dataset.send(((self,) + self.cp_dataset, None))

    def _updateDataset(self, widget, dataset):
        self.Outputs.dataset.send(((self, dataset) + self.cp_dataset[1:], widget))

    def setVisible(self, visible):
        pass



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "16/07/2021"


from silx.gui import qt
from silx.utils.enum import Enum as _Enum


class PixelSize(_Enum):
    """
    Different pixel sizes
    """

    Basler = 0.051
    PcoEdge_2x = 0.00375
    PcoEdge_10x = 0.00075


class RSMWidget(qt.QMainWindow):
    """
    Widget to transform axes of RSM data
    """

    sigComputed = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self._rotate = False
        self._moments = None
        self._pixelSize = None

        widget = qt.QWidget()
        layout = qt.QGridLayout()

        pixelSizeLabel = qt.QLabel("Pixel size: ")
        self._pixelSizeCB = qt.QComboBox()
        self._pixelSizeCB.addItems(PixelSize.names())
        self._rotateCB = qt.QCheckBox("Rotate RSM", self)
        self._okButton = qt.QPushButton("Ok")
        self._okButton.setEnabled(False)
        self._okButton.pressed.connect(self._saveRSM)
        layout.addWidget(pixelSizeLabel, 0, 0)
        layout.addWidget(self._pixelSizeCB, 0, 1)
        layout.addWidget(self._rotateCB, 1, 1)
        layout.addWidget(self._okButton, 2, 0, 1, 2)
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter.

        :param Dataset dataset: dataset
        """
        self.parent = parent
        self.dataset = dataset
        self.indices = indices
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset
        self._okButton.setEnabled(True)

    def getDataset(self):
        return self.dataset, self.indices, self.bg_indices, self.bg_dataset

    def _updateDataset(self, widget, dataset):
        self.parent._updateDataset(widget, dataset)
        self.dataset = dataset

    @property
    def pixelSize(self):
        return self._pixelSize

    @pixelSize.setter
    def pixelSize(self, pixelSize):
        self._pixelSize = pixelSize
        self._pixelSizeCB.setCurrentText(str(pixelSize))

    @property
    def rotate(self):
        return self._rotate

    @rotate.setter
    def rotate(self, rotate):
        self._rotate = rotate
        self._rotateCB.setChecked(rotate)

    def _saveRSM(self):
        self._pixelSize = self._pixelSizeCB.currentText()
        self._rotate = self._rotateCB.isChecked()
        self.dataset.compute_transformation(
            PixelSize[self._pixelSize].value, kind="rsm", rotate=self._rotate
        )
        self.sigComputed.emit()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "13/05/2020"

import numpy
from sklearn.decomposition import IncrementalPCA

from darfix.io import utils
from .base import Base


class IPCA(Base):
    """
    Compute PCA in chunks, using the Incremental principal component analysis implementation
    in scikit-learn.
    To compute W, partially fits the rows in chunks (reduced number of images). Then, to
    compute H, applies dimensionality reduction for every chunk, and horizontally stacks
    the projection into H.

    :param data: array of shape (n_samples, n_features). See `rowvar`.
    :type data: array_like
    :param chunksize: Size of every group of samples to apply PCA to. PCA will be fit with arrays
        of shape (chunksize, n_features), where nfeatures is the number of features per sample.
        Depending on `rowvar`, the chunks will be from the rows or from the columns.
    :type chunksize: int
    :param num_components: Number of components to keep, defaults to None.
    :type num_components: Union[None,int], optional
    :param whiten: If True, whitening is applied to the components.
    :type whiten: bool, optional
    :param indices: The indices of the samples to use, defaults to None. If `rowvar` is
        False, corresponds to the indices of the features to use.
    :type indices: Union[None,array_like], optional
    :param rowvar: If rowvar is True (default), then each row represents a sample,
        with features in the columns. Otherwise, the relationship is transposed: each
        column represents a sample, while the rows contain features.
    :type rowvar: bool, optional
    """

    def __init__(
        self,
        data,
        chunksize,
        num_components=None,
        whiten=False,
        indices=None,
        rowvar=True,
    ):
        Base.__init__(self, data, num_components=num_components, indices=indices)

        self._num_components = min(self._num_components, chunksize)

        self._chunksize = chunksize
        self._singular_values = None
        self._rowvar = rowvar
        self._ipca = IncrementalPCA(n_components=num_components, whiten=whiten)

    @property
    def singular_values(self):
        """
        The singular values corresponding to each of the selected components.

        :retuns: array, shape (n_components,)
        """
        return self._singular_values

    def _init_w(self):
        """
        Init W to None.
        """
        if self.W is None and self._rowvar:
            Base._init_w(self)
        elif not self._rowvar:
            self.W = numpy.random.random((self.num_features, self.num_components))

    def _init_h(self):
        if self.H is None and self._rowvar:
            Base._init_h(self)
        elif not self._rowvar:
            self.H = numpy.random.random((self.num_components, self.num_samples))

    def _update_h(self):
        """
        Matrix H is filled with the components of IPCA.
        """
        self.H = self._ipca.components_

    def _update_w(self):
        if self.indices is not None:
            if not self._rowvar:
                # Samples are in the columns
                # Partially fit every chunk
                utils.advancement_display(
                    0, self.num_features, "Fitting data with IPCA"
                )
                assert (
                    self._chunksize >= self.num_components
                ), "Chunksize has to be bigger than number of components"
                for i in range(0, self.num_features, self._chunksize):
                    if self.data.shape[1] - i < self.num_components:
                        # If at the last iteration the number of samples is smaller than
                        # the number of components.
                        utils.advancement_display(
                            self.num_features,
                            self.num_features,
                            "Fitting data with IPCA",
                        )
                        break
                    self._ipca.partial_fit(
                        self.data[self.indices, i : i + self._chunksize].T
                    )
                    utils.advancement_display(
                        i + self._chunksize, self.num_features, "Fitting data with IPCA"
                    )
                utils.advancement_display(
                    0, self.num_features, "Transforming data with IPCA"
                )
                # Transform every chunk to get W
                for i in range(0, self.num_features, self._chunksize):
                    self.W[i : i + self._chunksize] = self._ipca.transform(
                        self.data[self.indices, i : i + self._chunksize].T
                    )
                    utils.advancement_display(
                        i + self._chunksize,
                        self.num_features,
                        "Transforming data with IPCA",
                    )
            else:
                # Images are in the rows
                # Partially fit every chunk
                utils.advancement_display(
                    0, len(self.indices), "Fitting data with IPCA"
                )
                assert (
                    self._chunksize >= self.num_components
                ), "Chunksize has to be bigger than number of components"
                for i in range(0, len(self.indices), self._chunksize):
                    indx = self.indices[i : i + self._chunksize]
                    if len(indx) < self.num_components:
                        # If at the last iteration the number of samples is smaller than
                        # the number of components.
                        utils.advancement_display(
                            len(self.indices),
                            len(self.indices),
                            "Fitting data with IPCA",
                        )
                        break
                    self._ipca.partial_fit(self.data[indx])
                    utils.advancement_display(
                        i + self._chunksize, len(self.indices), "Fitting data with IPCA"
                    )
                utils.advancement_display(
                    0, len(self.indices), "Transforming data with IPCA"
                )
                # Transform every chunk to get W
                for i in range(0, len(self.indices), self._chunksize):
                    indx = self.indices[i : i + self._chunksize]
                    self.W[i : i + self._chunksize] = self._ipca.transform(
                        self.data[indx]
                    )
                    utils.advancement_display(
                        i + self._chunksize,
                        len(self.indices),
                        "Transforming data with IPCA",
                    )
        else:
            if not self._rowvar:
                # Samples are in the columns
                # Partially fit every chunk
                utils.advancement_display(
                    0, self.num_features, "Fitting data with IPCA"
                )
                assert (
                    self._chunksize >= self.num_components
                ), "Chunksize has to be bigger than number of components"
                for i in range(0, self.num_features, self._chunksize):
                    if self.data.shape[1] - i < self.num_components:
                        # If at the last iteration the number of samples is smaller than
                        # the number of components.
                        utils.advancement_display(
                            self.num_features,
                            self.num_features,
                            "Fitting data with IPCA",
                        )
                        break
                    self._ipca.partial_fit(self.data[:, i : i + self._chunksize].T)
                    utils.advancement_display(
                        i + self._chunksize, self.num_features, "Fitting data with IPCA"
                    )
                utils.advancement_display(
                    0, self.num_features, "Transforming data with IPCA"
                )
                # Transform every chunk to get W
                for i in range(0, self.num_features, self._chunksize):
                    self.W[i : i + self._chunksize] = self._ipca.transform(
                        self.data[:, i : i + self._chunksize].T
                    )
                    utils.advancement_display(
                        i + self._chunksize,
                        self.num_features,
                        "Transforming data with IPCA",
                    )
            else:
                # Partially fit every chunk
                utils.advancement_display(0, self.num_samples, "Fitting data with IPCA")
                assert (
                    self._chunksize >= self.num_components
                ), "Chunksize has to be bigger than number of components"
                for i in range(0, self.num_samples, self._chunksize):
                    if self.data.shape[0] - i < self.num_components:
                        # If at the last iteration the number of samples is smaller than
                        # the number of components.
                        utils.advancement_display(
                            self.num_samples, self.num_samples, "Fitting data with IPCA"
                        )
                        break
                    self._ipca.partial_fit(self.data[i : i + self._chunksize])
                    utils.advancement_display(
                        i + self._chunksize, self.num_samples, "Fitting data with IPCA"
                    )
                utils.advancement_display(
                    0, self.num_samples, "Transforming data with IPCA"
                )
                # Transform every chunk to get W
                for i in range(0, self.num_samples, self._chunksize):
                    self.W[i : i + self._chunksize] = self._ipca.transform(
                        self.data[i : i + self._chunksize]
                    )
                    utils.advancement_display(
                        i + self._chunksize,
                        self.num_samples,
                        "Transforming data with IPCA",
                    )

        self._singular_values = self._ipca.singular_values_

    def fit_transform(self, max_iter=1, error_step=None, W=None, H=None):

        self.W = None
        self.H = None

        Base.fit_transform(self, max_iter=max_iter, error_step=error_step)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "09/12/2020"

import os
import copy
import unittest
import numpy
import tempfile
import shutil

from silx.io.url import DataUrl

from darfix.test import utils
from darfix.core.dataset import Data


class TestDataset(unittest.TestCase):

    """Tests for class Dataset in `dataset.py`"""

    def setUp(self):
        self._dir = tempfile.mkdtemp()
        self.dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=3, header=True, _dir=self._dir
        )

    def test_data_load(self):
        """Tests the correct load of the data"""
        self.assertEqual(len(self.dataset.get_data()), 3)
        self.assertEqual(self.dataset.nframes, 3)

    def test_nframes(self):
        """Tests the nframes method"""
        self.assertEqual(self.dataset.nframes, 3)

    def test_deepcopy(self):
        """Tests the correct deepcopy of the object"""
        dataset_copy = copy.deepcopy(self.dataset)
        self.assertEqual(self.dataset.nframes, dataset_copy.nframes)
        self.assertEqual(self.dataset.data.shape, dataset_copy.data.shape)

    def test_zsum_in_memory(self):

        zsum = self.dataset.zsum()
        self.assertEqual(zsum.shape, (100, 100))

        result = numpy.sum(self.dataset.get_data(), axis=0)
        numpy.testing.assert_array_equal(zsum, result)

        indices = [0, 1]
        zsum = self.dataset.zsum(indices)
        result = numpy.sum(self.dataset.get_data(indices), axis=0)
        numpy.testing.assert_array_equal(zsum, result)

    def test_zsum_in_disk(self):

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )
        indices = [0, 1, 2, 3, 4]
        zsum = dataset.zsum(indices)
        self.assertEqual(zsum.shape, (100, 100))
        data = []
        for i in indices:
            data += [dataset.get_data(i)]
        result = numpy.sum(data, axis=0)
        numpy.testing.assert_array_equal(zsum, result)

    def test_filter_data(self):
        """Tests the correct separation of empty frames and data frames"""
        dims = (10, 100, 100)
        data = numpy.zeros(dims)
        background = numpy.random.random(dims)
        idxs = [0, 2, 4]
        data[idxs] += background[idxs]
        dataset = utils.createDataset(data=data, _dir=self._dir)
        used_idx, not_used_idx = dataset.partition_by_intensity(bottom_bin=1)
        self.assertEqual(used_idx.shape[0], 3)
        self.assertEqual(not_used_idx.shape[0], 7)

    def test_bs_in_memory(self):
        """Tests the background subtraction function with data in memory"""
        indices = [0, 1]
        bs_dataset = self.dataset.apply_background_subtraction(
            background=[2], indices=indices
        )

        self.assertEqual(bs_dataset.nframes, self.dataset.nframes)
        self.assertNotEqual(bs_dataset.data.urls[0], self.dataset.data.urls[0])
        numpy.testing.assert_equal(self.dataset.data.metadata, bs_dataset.data.metadata)
        numpy.testing.assert_equal(self.dataset.data[2], bs_dataset.data[2])

    def test_bs_in_disk(self):
        """Tests the background subtraction function with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )
        indices = [0, 2, 4, 5, 7]
        bs_dataset = dataset.apply_background_subtraction(
            background=[1, 3, 6, 8, 9], indices=indices
        )

        self.assertEqual(bs_dataset.nframes, dataset.nframes)
        self.assertNotEqual(dataset.data.urls[0], bs_dataset.data.urls[0])
        numpy.testing.assert_equal(dataset.data[6], bs_dataset.data[6])

    def test_bs_mean_in_disk(self):
        """Tests the correct subtraction of the background with data in disk and mean method"""
        data = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [2, 2, 3, 4, 5],
                    [3, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 3],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [8, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
            ]
        )

        dark = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
            ]
        )

        expected = numpy.array(
            [
                [
                    [0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 0],
                    [2, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [7, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
            ]
        )

        dataset = utils.createDataset(data, in_memory=False)

        # Background data in memory

        dark_dataset = utils.createDataset(dark, in_memory=True)
        new_dataset = dataset.apply_background_subtraction(
            background=dark_dataset, method="mean"
        )

        new_dataset.in_memory = True
        numpy.testing.assert_array_equal(expected, new_dataset.data)

        # Background data in disk

        dark_dataset = utils.createDataset(dark, in_memory=False)

        new_dataset = dataset.apply_background_subtraction(
            background=dark_dataset, method="mean"
        )
        new_dataset.in_memory = True
        numpy.testing.assert_array_equal(expected, new_dataset.data)

    def test_bs_median_in_disk(self):
        """Tests the correct subtraction of the background with data in disk and median method"""
        data = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [2, 2, 3, 4, 5],
                    [3, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 3],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [8, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
            ]
        )

        dark = numpy.array(
            [
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
                [
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                    [1, 2, 3, 4, 5],
                ],
            ]
        )

        expected = numpy.array(
            [
                [
                    [0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 0],
                    [2, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [7, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
            ]
        )

        dataset = utils.createDataset(data, in_memory=False)

        # Background data in memory

        dark_dataset = utils.createDataset(dark, in_memory=True)

        new_dataset = dataset.apply_background_subtraction(
            background=dark_dataset, method="median"
        )
        new_dataset.in_memory = True
        numpy.testing.assert_array_equal(expected, new_dataset.data)

        # Background data in disk

        dark_dataset = utils.createDataset(dark, in_memory=False)

        new_dataset = dataset.apply_background_subtraction(
            background=dark_dataset, method="median"
        )
        new_dataset.in_memory = True
        numpy.testing.assert_array_equal(expected, new_dataset.data)

        # Compute background with chunks
        new_dataset = dataset.apply_background_subtraction(
            background=dark_dataset, method="median", chunk_shape=[2, 2]
        )
        new_dataset.in_memory = True
        numpy.testing.assert_array_equal(expected, new_dataset.data)

        # Compute background with step
        new_dataset = dataset.apply_background_subtraction(
            background=dark_dataset, method="median", step=2
        )
        new_dataset.in_memory = True
        numpy.testing.assert_array_equal(expected, new_dataset.data)

    def test_hp_in_memory(self):
        """Tests the hot pixel removal function with data in memory"""
        indices = [0, 1]
        hp_dataset = self.dataset.apply_hot_pixel_removal(indices=indices)

        self.assertEqual(hp_dataset.nframes, self.dataset.nframes)
        self.assertNotEqual(hp_dataset.data.urls[0], self.dataset.data.urls[0])
        numpy.testing.assert_equal(self.dataset.data.metadata, hp_dataset.data.metadata)
        numpy.testing.assert_equal(self.dataset.data[2], hp_dataset.data[2])

    def test_hp_in_disk(self):
        """Tests the hot pixel removal function with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )
        indices = [0, 1, 2, 3, 4]
        hp_dataset = dataset.apply_hot_pixel_removal(indices=indices)

        self.assertEqual(hp_dataset.nframes, dataset.nframes)
        self.assertNotEqual(hp_dataset.data.urls[0], dataset.data.urls[0])
        numpy.testing.assert_equal(dataset.data[5], hp_dataset.data[5])

    def test_threshold_in_memory(self):
        """Tests the threshold removal function with data in memory"""
        indices = [0, 1]
        threshold_dataset = self.dataset.apply_threshold_removal(indices=indices)

        self.assertEqual(threshold_dataset.nframes, self.dataset.nframes)
        self.assertNotEqual(threshold_dataset.data.urls[0], self.dataset.data.urls[0])
        numpy.testing.assert_equal(
            self.dataset.data.metadata, threshold_dataset.data.metadata
        )
        numpy.testing.assert_equal(self.dataset.data[2], threshold_dataset.data[2])

    def test_threshold_in_disk(self):
        """Tests the threshold removal function with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )
        indices = [0, 1, 2, 3, 4]
        threshold_dataset = dataset.apply_threshold_removal(indices=indices)

        self.assertEqual(threshold_dataset.nframes, dataset.nframes)
        self.assertNotEqual(threshold_dataset.data.urls[0], dataset.data.urls[0])
        numpy.testing.assert_equal(dataset.data[5], threshold_dataset.data[5])

    def test_roi_in_memory(self):
        """Tests the roi function with data in memory"""
        new_dataset = self.dataset.apply_roi(origin=[0, 0], size=[20, 20])
        self.assertEqual(new_dataset.nframes, 3)
        numpy.testing.assert_equal(self.dataset.data[:, :20, :20], new_dataset.data)

    def test_roi_in_disk(self):
        """Tests the roi function with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )
        indices = [0, 1, 2, 3, 4]
        new_dataset = dataset.apply_roi(origin=[0, 0], size=[20, 20], indices=indices)

        self.assertEqual(new_dataset.nframes, 5)
        numpy.testing.assert_equal(
            dataset.get_data(indices)[0][:20, :20], new_dataset.data[0]
        )

    def test_roi_bs_in_memory(self):
        """Tests the background subtraction function with data in disk"""

        roi_dataset = self.dataset.apply_roi(origin=[0, 0], size=[20, 20])
        indices = [0, 1]
        bs_dataset = roi_dataset.apply_background_subtraction(
            background=[2], indices=indices
        )

        self.assertEqual(bs_dataset.nframes, self.dataset.nframes)
        self.assertNotEqual(self.dataset.data.urls[0], bs_dataset.data.urls[0])
        numpy.testing.assert_equal(bs_dataset.data[2], roi_dataset.data[2])

    def test_roi_bs_in_disk(self):
        """Tests the background subtraction function with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )
        roi_dataset = dataset.apply_roi(origin=[0, 0], size=[20, 20])
        indices = [0, 2, 4, 5, 7]
        bs_dataset = roi_dataset.apply_background_subtraction(
            background=[1, 3, 6, 8, 9], indices=indices
        )
        self.assertEqual(bs_dataset.nframes, dataset.nframes)
        self.assertNotEqual(dataset.data.urls[0], bs_dataset.data.urls[0])
        numpy.testing.assert_equal(bs_dataset.data[6], roi_dataset.data[6])

    def test_roi_hp_in_memory(self):
        """Tests the background subtraction function with data in disk"""

        roi_dataset = self.dataset.apply_roi(origin=[0, 0], size=[20, 20])
        indices = [0, 1]
        hp_dataset = roi_dataset.apply_hot_pixel_removal(indices=indices)

        self.assertEqual(hp_dataset.nframes, self.dataset.nframes)
        self.assertNotEqual(hp_dataset.data.urls[0], self.dataset.data.urls[0])
        numpy.testing.assert_equal(self.dataset.data.metadata, hp_dataset.data.metadata)
        numpy.testing.assert_equal(roi_dataset.data[2], hp_dataset.data[2])

    def test_roi_hp_in_disk(self):
        """Tests the background subtraction function with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )
        roi_dataset = dataset.apply_roi(origin=[0, 0], size=[20, 20])
        indices = [0, 2, 4, 5, 7]
        hp_dataset = roi_dataset.apply_hot_pixel_removal(indices=indices)
        self.assertEqual(hp_dataset.nframes, dataset.nframes)
        self.assertNotEqual(dataset.data.urls[0], hp_dataset.data.urls[0])
        numpy.testing.assert_equal(hp_dataset.data[6], roi_dataset.data[6])

    def test_roi_threshold_in_memory(self):
        """Tests the threshold removal function with data in memory"""
        roi_dataset = self.dataset.apply_roi(origin=[0, 0], size=[20, 20])
        indices = [0, 1]
        threshold_dataset = roi_dataset.apply_threshold_removal(indices=indices)

        self.assertEqual(threshold_dataset.nframes, self.dataset.nframes)
        self.assertNotEqual(threshold_dataset.data.urls[0], self.dataset.data.urls[0])
        numpy.testing.assert_equal(
            self.dataset.data.metadata, threshold_dataset.data.metadata
        )
        numpy.testing.assert_equal(roi_dataset.data[2], threshold_dataset.data[2])

    def test_roi_threshold_in_disk(self):
        """Tests the threshold removal function with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )
        roi_dataset = dataset.apply_roi(origin=[0, 0], size=[20, 20])
        indices = [0, 1, 2, 3, 4]
        threshold_dataset = roi_dataset.apply_threshold_removal(indices=indices)

        self.assertEqual(threshold_dataset.nframes, dataset.nframes)
        self.assertNotEqual(threshold_dataset.data.urls[0], dataset.data.urls[0])
        numpy.testing.assert_equal(roi_dataset.data[5], threshold_dataset.data[5])

    def test_find_shift(self):
        """Tests the shift detection"""
        shift = self.dataset.find_shift()

        self.assertEqual(shift.shape, (2, 3))

        shift = self.dataset.find_shift(indices=[0, 1])

        self.assertEqual(shift.shape, (2, 2))

    def test_apply_shift(self):
        """Tests the shift correction"""

        new_dataset = self.dataset.apply_shift(
            shift=numpy.array([[0, 0.5, 1], [0, 0.5, 1]])
        )

        self.assertEqual(new_dataset.nframes, self.dataset.nframes)
        self.assertNotEqual(new_dataset.data.urls[0], self.dataset.data.urls[0])

    def test_pca_in_memory(self):
        """Tests PCA with data in memory"""

        H, W = self.dataset.pca()

        self.assertEqual(
            H.shape, (self.dataset.nframes, len(self.dataset.data[0].flatten()))
        )
        self.assertEqual(W.shape, (self.dataset.nframes, self.dataset.nframes))

        n_components = 1
        indices = [1, 2]
        H, W = self.dataset.pca(num_components=n_components, indices=indices)

        self.assertEqual(H.shape, (n_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (2, n_components))

    def test_pca_in_disk(self):
        """Tests PCA with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )

        H, W = dataset.pca(chunk_size=1000)

        self.assertEqual(H.shape, (dataset.nframes, len(dataset.data[0].flatten())))
        self.assertEqual(W.shape, (dataset.nframes, dataset.nframes))

        n_components = 3
        indices = [0, 1, 3, 4, 6]
        H, W = dataset.pca(
            chunk_size=1000, num_components=n_components, indices=indices
        )

        self.assertEqual(H.shape, (n_components, len(dataset.data[0].flatten())))
        self.assertEqual(W.shape, (5, n_components))

    def test_nmf_in_memory(self):
        """Tests NMF with data in memory"""

        num_components = 2
        H, W = self.dataset.nmf(num_components)

        self.assertEqual(H.shape, (num_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (self.dataset.nframes, num_components))

        num_components = 1
        indices = [1, 2]
        H, W = self.dataset.nmf(num_components=num_components, indices=indices)

        self.assertEqual(H.shape, (num_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (2, num_components))

    def test_nmf_in_disk(self):
        """Tests NMF with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, in_memory=False
        )

        num_components = 5
        H, W = dataset.nmf(
            num_components=num_components, num_iter=10, waterfall=[100, 50, 25]
        )

        self.assertEqual(H.shape, (num_components, len(dataset.data[0].flatten())))
        self.assertEqual(W.shape, (dataset.nframes, num_components))

        n_components = 4
        indices = numpy.random.choice(10, size=8, replace=False)
        H, W = dataset.nmf(num_components=n_components, indices=indices)

        self.assertEqual(H.shape, (n_components, len(dataset.data[0].flatten())))
        self.assertEqual(W.shape, (len(indices), n_components))

    def test_nica_in_memory(self):
        """Tests NICA with data in memory"""

        num_components = 2
        H, W = self.dataset.nica(num_components)

        self.assertEqual(H.shape, (num_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (self.dataset.nframes, num_components))

        num_components = 1
        indices = [1, 2]
        H, W = self.dataset.nica(num_components=num_components, indices=indices)

        self.assertEqual(H.shape, (num_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (2, num_components))

    def test_nica_in_disk(self):
        """Tests NICA with data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=100, in_memory=False
        )

        num_components = 5
        H, W = dataset.nica(num_components=num_components, chunksize=1000)

        self.assertEqual(H.shape, (num_components, len(dataset.data[0].flatten())))
        self.assertEqual(W.shape, (dataset.nframes, num_components))

        n_components = 10
        indices = numpy.random.choice(100, size=80, replace=False)
        H, W = dataset.nica(
            num_components=n_components, indices=indices, chunksize=1000
        )

        self.assertEqual(H.shape, (n_components, len(dataset.data[0].flatten())))
        self.assertEqual(W.shape, (len(indices), n_components))

    def test_apply_fit_in_memory(self):
        """Tests fitting data in memory"""
        dataset = utils.createRandomDataset(
            dims=(10, 100), nb_data_files=10, header=True, _dir=self._dir
        )
        new_dataset, maps = dataset.apply_fit()

        self.assertEqual(new_dataset.nframes, dataset.nframes)
        self.assertNotEqual(new_dataset.data.urls[0], self.dataset.data.urls[0])
        self.assertEqual(len(maps), 4)

    def test_apply_fit_in_disk(self):
        """Tests fitting data in disk"""

        dataset = utils.createRandomDataset(
            dims=(10, 100), nb_data_files=10, in_memory=False
        )
        indices = [0, 1, 2, 3, 4]
        fit_dataset, maps = dataset.apply_fit(indices=indices)

        self.assertEqual(fit_dataset.nframes, dataset.nframes)
        self.assertNotEqual(fit_dataset.data.urls[0], dataset.data.urls[0])
        numpy.testing.assert_equal(fit_dataset.data[5], dataset.data[5])
        self.assertEqual(len(maps), 4)

    def test_compute_rsm_in_memory(self):
        """Tests fitting data in memory"""
        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, header=True, _dir=self._dir
        )
        dataset.compute_transformation(d=0.1, kind="rsm")

        self.assertEqual(dataset.transformation.shape, dataset.get_data(0).shape)

    def test_apply_rsm_in_disk(self):
        """Tests fitting data in disk"""

        dataset = utils.createRandomDataset(
            dims=(100, 100), nb_data_files=10, header=True, in_memory=False
        )
        dataset.compute_transformation(d=0.1, kind="rsm")

        self.assertEqual(dataset.transformation.shape, dataset.get_data(0).shape)


class TestData(unittest.TestCase):

    """Tests for class Data in `dataset.py`"""

    def setUp(self):
        self._dir = tempfile.mkdtemp()
        self.urls = []
        self.metadata = []
        self.data = []
        for i in range(10):
            url = self._dir + str(i) + ".npy"
            data = numpy.random.rand(100, 100)
            self.data += [data]
            self.urls += [DataUrl(file_path=url, scheme="fabio")]
            self.metadata += ["No metadata"]
            numpy.save(url, data)

        self.data = numpy.array(self.data)

    def test_create_data_in_memory(self):
        """Tests the correct creation of the data in memory"""
        data = Data(urls=self.urls, metadata=self.metadata)

        for i in range(10):
            self.assertEqual(data.urls[i], self.urls[i])
            self.assertEqual(data.metadata[i], self.metadata[i])

        numpy.testing.assert_array_equal(self.data, data)

    def test_create_data_in_disk(self):
        """Tests the correct creation of the data in disk"""
        data = Data(urls=self.urls, metadata=self.metadata, in_memory=False)

        for i in range(10):
            self.assertEqual(data.urls[i], self.urls[i])
            self.assertEqual(data.metadata[i], self.metadata[i])

    def test_get_in_memory(self):
        """Tests the correct get of the data in memory"""
        data = Data(urls=self.urls, metadata=self.metadata)

        for i in range(10):
            numpy.testing.assert_array_equal(data[i], self.data[i])

        indices = [0, 2, 6]
        numpy.testing.assert_array_equal(self.data[indices], data[indices])

    def test_get_in_disk(self):
        """Tests the correct get of the data in disk"""
        data = Data(urls=self.urls, metadata=self.metadata, in_memory=False)

        for i in range(10):
            numpy.testing.assert_array_equal(data[i], self.data[i])

        indices = [0, 2, 6]
        for i in range(3):
            numpy.testing.assert_array_equal(self.data[indices][i], data[indices][i])

    def test_get_slices_in_memory(self):
        """Tests getting slices of the data in memory"""
        data = Data(urls=self.urls, metadata=self.metadata)

        numpy.testing.assert_array_equal(data[2:3, 50:, 50:], self.data[2:3, 50:, 50:])

    def test_get_slices_in_disk(self):
        """Tests getting slices of the data in disk"""
        data = Data(urls=self.urls, metadata=self.metadata, in_memory=False)

        for i in range(3):
            numpy.testing.assert_array_equal(data[2:5][i], self.data[2:5][i])

    def test_shape_in_memory(self):
        """Tests the correct shape of the data in memory"""
        data = Data(urls=self.urls, metadata=self.metadata)
        self.assertEqual(data.shape, self.data.shape)

        urls = numpy.array(self.urls).reshape((5, 2))
        data = Data(urls=urls, metadata=self.metadata)
        self.assertEqual(data.shape, (5, 2, 100, 100))

    def test_shape_in_disk(self):
        """Tests the correct shape of the data in disk"""
        data = Data(urls=self.urls, metadata=self.metadata, in_memory=False)
        self.assertEqual(data.shape, self.data.shape)

        urls = numpy.array(self.urls).reshape((5, 2))
        data = Data(urls=urls, metadata=self.metadata, in_memory=False)
        self.assertEqual(data.shape, (5, 2, 100, 100))

    def test_save(self):
        data = numpy.random.rand(10, 100, 100).view(Data)
        # data.in_memory = True
        data.save(os.path.join(self._dir, "data.hdf5"))

        numpy.testing.assert_array_equal(data.shape, (10, 100, 100))

    def test_save_new_shape(self):
        data = numpy.random.rand(10, 50, 50).view(Data)
        # data.in_memory = True
        data.save(os.path.join(self._dir, "data.hdf5"), new_shape=(10, 50, 50))

        numpy.testing.assert_array_equal(data.shape, (10, 50, 50))

    def test_convert_to_hdf5(self):
        """Tests the reshape of the data"""
        data = Data(urls=self.urls, metadata=self.metadata)
        dataset = data.convert_to_hdf5(self._dir)
        self.assertEqual(dataset.shape, (10, 10000))
        data._file.close()

        data = Data(urls=self.urls, metadata=self.metadata, in_memory=False)
        dataset = data.convert_to_hdf5(self._dir)
        self.assertEqual(dataset.shape, (10, 10000))

    def test_reshape(self):
        """Tests the reshape of the data"""
        data = Data(urls=self.urls, metadata=self.metadata)
        new_data = data.reshape((2, 5, 100, 100))
        self.assertEqual(new_data.shape, (2, 5, 100, 100))

        data = Data(urls=self.urls, metadata=self.metadata, in_memory=False)
        new_data = data.reshape((2, 5, 100, 100))
        self.assertEqual(new_data.shape, (2, 5, 100, 100))

    def test_take(self):
        """Tests the take of the data"""
        data = Data(urls=self.urls, metadata=self.metadata)
        result = numpy.take(self.data, [0, 1, 2, 3], axis=0)
        new_data = data.take([0, 1, 2, 3])
        numpy.testing.assert_array_equal(new_data, result)

        data = Data(urls=self.urls, metadata=self.metadata, in_memory=False)
        new_data = data.take([0, 1, 2, 3])
        for i in range(4):
            numpy.testing.assert_array_equal(new_data[i], result[i])

    def test_flatten(self):
        """Tests the flatten of the data"""
        urls = numpy.array(self.urls).reshape((2, 5))
        data = Data(urls=urls, metadata=self.metadata)
        numpy.testing.assert_array_equal(self.data, data.flatten())

        data = Data(urls=urls, metadata=self.metadata, in_memory=False)
        for i in range(4):
            numpy.testing.assert_array_equal(data.flatten()[i], self.data[i])

    def tearDown(self):
        shutil.rmtree(self._dir)


if __name__ == "__main__":
    unittest.main()



import os
import sys
import glob
from pprint import pformat
from ewoks import load_graph
from ewoks import execute_graph as _execute_graph
from darfix.core.process import graph_data_selection


def is_image_file(filename):
    if not os.path.isfile(filename):
        return False
    return all(not filename.endswith(ext) for ext in [".py", ".ows"])


def execute_graph(argv=None, **extra_options):
    import argparse

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="Execute a darfix workflow", prog="darfix"
    )
    parser.add_argument("-wf", "--workflow", help="Filename of the workflow", type=str)
    parser.add_argument(
        "-fd",
        "--file_directory",
        help="Directory containing images",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-ff",
        "--first_filename",
        help="Filename to the first file of the stack",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-td",
        "--treated_data",
        help="Directory to save treated data",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--in-disk",
        help="Do not load data into memory",
        dest="in_disk",
        action="store_true",
        default=None,
    )

    args = parser.parse_args(argv[1:])
    if not args.workflow:
        parser.error("Please enter the workflow filename")
    if not (args.file_directory or args.first_filename):
        parser.error("Please enter the file directory or first filename")

    root_dir = args.file_directory
    if args.first_filename:
        filenames = args.first_filename
    else:
        filenames = sorted(
            [x for x in glob.glob(os.path.join(root_dir, "*")) if is_image_file(x)]
        )
    if args.treated_data:
        root_dir = args.treated_data
    in_memory = not args.in_disk

    print("\nLoading workflow", repr(args.workflow), "...")
    graph = load_graph(args.workflow)

    print("Setting workflow inputs parameters ...")
    graph_data_selection(
        graph=graph, filenames=filenames, root_dir=root_dir, in_memory=in_memory
    )

    print("Executing workflow ...")
    results = _execute_graph(graph, **extra_options)
    print("Result of workflow '%s':\n%s" % (args.workflow, pformat(results)))
    print("Finished\n")

    return results


def main(argv=None):
    execute_graph(argv=argv)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "10/08/2021"

from matplotlib.colors import hsv_to_rgb
import numpy
import logging

from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.image.marchingsquares import find_contours
from silx.math.medianfilter import medfilt2d
from silx.utils.enum import Enum as _Enum
from silx.io.dictdump import dicttonx

from darfix.io.utils import create_nxdata_dict
from darfix.gui.utils import ChooseDimensionWidget
from .operationThread import OperationThread

_logger = logging.getLogger(__file__)


class Method(_Enum):
    """
    Different maps to show
    """

    COM = "Center of mass"
    FWHM = "FWHM"
    SKEWNESS = "Skewness"
    KURTOSIS = "Kurtosis"
    ORI_DIST = "Orientation distribution"
    MOSAICITY = "Mosaicity"


class GrainPlotWidget(qt.QMainWindow):
    """
    Widget to show a series of maps for the analysis of the data.
    """

    sigComputed = qt.Signal()

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        self._methodCB = qt.QComboBox()
        self._methodCB.addItems(Method.values())
        for i in range(len(Method)):
            self._methodCB.model().item(i).setEnabled(False)
        self._methodCB.currentTextChanged.connect(self._updatePlot)
        self._plotWidget = qt.QWidget()
        plotsLayout = qt.QHBoxLayout()
        self._plotWidget.setLayout(plotsLayout)
        self._contoursPlot = Plot2D(parent=self)
        widget = qt.QWidget(parent=self)
        layout = qt.QVBoxLayout()
        self._levelsWidget = qt.QWidget()
        levelsLayout = qt.QGridLayout()
        levelsLabel = qt.QLabel("Number of levels:")
        self._levelsLE = qt.QLineEdit("20")
        self._levelsLE.setToolTip("Number of levels to use when finding the contours")
        self._levelsLE.setValidator(qt.QIntValidator())
        self._computeContoursB = qt.QPushButton("Compute")
        self._motorValuesCheckbox = qt.QCheckBox("Use motor values")
        self._motorValuesCheckbox.setChecked(True)
        self._motorValuesCheckbox.stateChanged.connect(self._checkboxStateChanged)
        self._centerDataCheckbox = qt.QCheckBox("Center angle values")
        self._centerDataCheckbox.setEnabled(False)
        self._centerDataCheckbox.stateChanged.connect(self._checkboxStateChanged)
        self._chooseDimensionWidget = ChooseDimensionWidget(
            self, vertical=False, values=False, _filter=False
        )
        self._chooseDimensionWidget.filterChanged.connect(self._updateMotorAxis)
        self._thirdMotorCB = qt.QComboBox(self)
        self._thirdMotorCB.currentIndexChanged.connect(self._updateThirdMotor)
        levelsLayout.addWidget(levelsLabel, 0, 0, 1, 1)
        levelsLayout.addWidget(self._levelsLE, 0, 1, 1, 1)
        levelsLayout.addWidget(self._motorValuesCheckbox, 0, 2, 1, 1)
        levelsLayout.addWidget(self._centerDataCheckbox, 0, 3, 1, 1)
        levelsLayout.addWidget(self._thirdMotorCB, 1, 2, 1, 1)
        levelsLayout.addWidget(self._computeContoursB, 1, 3, 1, 1)
        levelsLayout.addWidget(self._contoursPlot, 2, 0, 1, 4)
        self._levelsWidget.setLayout(levelsLayout)
        self._levelsWidget.hide()
        self._mosaicityPlot = Plot2D(parent=self)
        self._exportButton = qt.QPushButton("Export maps")
        self._exportButton.setEnabled(False)
        self._exportButton.clicked.connect(self.exportMaps)
        layout.addWidget(self._methodCB)
        layout.addWidget(self._chooseDimensionWidget)
        layout.addWidget(self._levelsWidget)
        layout.addWidget(self._plotWidget)
        layout.addWidget(self._mosaicityPlot)
        layout.addWidget(self._exportButton)
        self._plotWidget.hide()
        self._thirdMotorCB.hide()
        self._chooseDimensionWidget.hide()
        self._mosaicityPlot.hide()
        self._mosaicityPlot.getColorBarWidget().hide()
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter.

        :param Dataset dataset: dataset
        """
        self._parent = parent
        self.dataset = dataset
        self.indices = indices
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset
        self._mosaicity = None
        self._moments = None
        self.dimensions = [0, 1]
        state = self._methodCB.blockSignals(True)
        for i in range(len(Method)):
            self._methodCB.model().item(i).setEnabled(False)
        self._methodCB.blockSignals(state)
        if self.dataset.dims.ndim > 1 and self.dataset.dims.ndim < 4:
            self._curves = {}
            if self.dataset.dims.ndim == 3:
                # Update chooseDimensionWidget and thirdMotorCB
                state = self._chooseDimensionWidget.blockSignals(True)
                self._chooseDimensionWidget.setDimensions(self.dataset.dims)
                self._chooseDimensionWidget._updateState(True)
                self._chooseDimensionWidget.blockSignals(state)
                self._thirdMotorCB.clear()
                third_motor = self.dataset.dims.get(
                    int(
                        numpy.setdiff1d(
                            range(3), self._chooseDimensionWidget.dimension
                        )[0]
                    )
                )
                state = self._thirdMotorCB.blockSignals(True)
                self._thirdMotorCB.addItems(
                    numpy.array(third_motor.unique_values, dtype=str)
                )
                self._thirdMotorCB.blockSignals(state)
            self.ori_dist, self.hsv_key = self.dataset.compute_mosaicity_colorkey(
                third_motor=0
            )
            self._checkboxStateChanged()
            self._contoursPlot.getColorBarWidget().hide()
            self._curvesColormap = Colormap(
                name="temperature",
                vmin=numpy.min(self.ori_dist),
                vmax=numpy.max(self.ori_dist),
            )
            self._computeContoursB.clicked.connect(self._computeContours)
            self._methodCB.model().item(4).setEnabled(True)
            self._methodCB.setCurrentIndex(4)
        self._thread = OperationThread(self, self.dataset.apply_moments)
        self._thread.setArgs(self.indices)
        self._thread.finished.connect(self._updateData)
        self._thread.start()
        for i in reversed(range(self._plotWidget.layout().count())):
            self._plotWidget.layout().itemAt(i).widget().setParent(None)

        self._contoursPlot.setGraphTitle(
            self.dataset.title + "\n" + Method.ORI_DIST.value
        )
        self._mosaicityPlot.setGraphTitle(
            self.dataset.title + "\n" + Method.MOSAICITY.value
        )
        self._plots = []
        for axis, dim in self.dataset.dims:
            self._plots += [Plot2D(parent=self)]
            self._plots[-1].setGraphTitle(self.dataset.title + "\n" + dim.name)
            self._plots[-1].setDefaultColormap(Colormap(name="viridis"))
            self._plotWidget.layout().addWidget(self._plots[-1])

    def _updateData(self):
        """
        Updates the plots with the data computed in the thread
        """
        self._thread.finished.disconnect(self._updateData)
        if self._thread.data is not None:
            self._moments = self._thread.data
            self._updatePlot(self._methodCB.currentText())
            rg = len(Method) if self.dataset.dims.ndim > 1 else 4
            for i in range(rg):
                self._methodCB.model().item(i).setEnabled(True)
            self.mosaicity = self.compute_mosaicity()
            self._methodCB.setCurrentIndex(0)
            self._exportButton.setEnabled(True)
        else:
            print("\nComputation aborted")
            msg = qt.QMessageBox()
            msg.setIcon(qt.QMessageBox.Critical)
            msg.setText("Grain plot not working without dimensions")
            msg.setWindowTitle("Don't eat yellow snow!")
            msg.setStandardButtons(qt.QMessageBox.Ok)
            msg.exec_()

    def _updateDataset(self, widget, dataset):
        self._parent._updateDataset(widget, dataset)
        self.dataset = dataset

    def _checkboxStateChanged(self, state=None):
        """
        Update widgets linked to the checkbox state
        """
        self._centerDataCheckbox.setEnabled(not self._motorValuesCheckbox.isChecked())
        self._contoursPlot.remove(kind="curve")
        self._contoursPlot.resetZoom()
        scale = 100
        xdim = self.dataset.dims.get(self.dimensions[1])
        ydim = self.dataset.dims.get(self.dimensions[0])
        xscale = xdim.range[2]
        yscale = ydim.range[2]
        if self._motorValuesCheckbox.isChecked():
            self.ori_dist_origin = (xdim.range[0], ydim.range[0])
        elif self._centerDataCheckbox.isChecked():
            self.ori_dist_origin = (
                -xscale * int(xdim.size / 2),
                -yscale * int(ydim.size / 2),
            )
        else:
            self.ori_dist_origin = (0, 0)
        self._contoursPlot.addImage(
            hsv_to_rgb(self.hsv_key),
            xlabel=xdim.name,
            ylabel=ydim.name,
            origin=self.ori_dist_origin,
            scale=(xscale / scale, yscale / scale),
        )

    def _computeContours(self):
        """
        Compute contours map based on orientation distribution.
        """
        self._contoursPlot.remove(kind="curve")

        if self.ori_dist is not None:
            polygons = []
            levels = []
            for i in numpy.linspace(
                numpy.min(self.ori_dist),
                numpy.max(self.ori_dist),
                int(self._levelsLE.text()),
            ):
                polygons.append(find_contours(self.ori_dist, i))
                levels.append(i)

            colors = self._curvesColormap.applyToData(levels)
            xdim = self.dataset.dims.get(self.dimensions[1])
            ydim = self.dataset.dims.get(self.dimensions[0])
            self._curves = {}
            for ipolygon, polygon in enumerate(polygons):
                # iso contours
                for icontour, contour in enumerate(polygon):
                    if len(contour) == 0:
                        continue
                    # isClosed = numpy.allclose(contour[0], contour[-1])
                    x = contour[:, 1]
                    y = contour[:, 0]
                    xscale = xdim.range[2]
                    yscale = ydim.range[2]
                    x = self.ori_dist_origin[0] + x * xscale + xscale / 2
                    y = self.ori_dist_origin[1] + y * yscale + yscale / 2
                    legend = "poly{}.{}".format(icontour, ipolygon)
                    self._curves[legend] = {
                        "points": (x.copy(), y.copy()),
                        "color": colors[ipolygon],
                    }
                    self._contoursPlot.addCurve(
                        x=x,
                        y=y,
                        linestyle="-",
                        linewidth=2.0,
                        legend=legend,
                        resetzoom=False,
                        color=colors[ipolygon],
                    )

    def compute_mosaicity(self):
        """
        Compute mosaicity map depending on the selected dimensions, if any.
        """
        if self._moments is not None and self.dataset.dims.ndim > 1:
            com0 = self._moments[self.dimensions[0]][0]
            no_nans_indices = ~numpy.isnan(com0)
            min_com0 = numpy.min(
                com0[no_nans_indices] if len(com0[no_nans_indices]) > 0 else com0
            )
            com0[numpy.isnan(com0)] = min_com0
            norms0 = (com0 - min_com0) / numpy.ptp(com0)
            com1 = self._moments[self.dimensions[1]][0]
            no_nans_indices = ~numpy.isnan(com1)
            min_com1 = numpy.min(
                com1[no_nans_indices] if len(com1[no_nans_indices]) > 0 else com1
            )
            com1[numpy.isnan(com1)] = min_com1
            norms1 = (com1 - min_com1) / numpy.ptp(com1)

            mosaicity = hsv_to_rgb(
                numpy.stack(
                    (norms0, norms1, numpy.ones(self._moments[0].shape[1:])), axis=2
                )
            )
            return mosaicity

    def _updatePlot(self, method):
        """
        Update shown plots in the widget
        """
        method = Method(method)
        self._levelsWidget.hide()
        self._mosaicityPlot.hide()
        self._thirdMotorCB.hide()
        self._chooseDimensionWidget.hide()
        if method == Method.ORI_DIST:
            self._levelsWidget.show()
            self._plotWidget.hide()
            if self.dataset.dims.ndim == 3:
                self._chooseDimensionWidget.show()
                self._thirdMotorCB.show()
        elif method == Method.FWHM:
            self._plotWidget.show()
            for i, plot in enumerate(self._plots):
                self._addImage(plot, self._moments[i][1])
        elif method == Method.COM:
            self._plotWidget.show()
            for i, plot in enumerate(self._plots):
                self._addImage(plot, self._moments[i][0])
        elif method == Method.SKEWNESS:
            self._plotWidget.show()
            for i, plot in enumerate(self._plots):
                self._addImage(plot, self._moments[i][2])
        elif method == Method.KURTOSIS:
            self._plotWidget.show()
            for i, plot in enumerate(self._plots):
                self._addImage(plot, self._moments[i][3])
        elif method == Method.MOSAICITY:
            try:
                self._plotWidget.hide()
                self._addImage(self._mosaicityPlot, self.mosaicity)
                self._mosaicityPlot.show()
                if self.dataset.dims.ndim == 3:
                    self._chooseDimensionWidget.show()
            except Exception as e:
                _logger.error("Couldn't compute mosaicity: ", e)
        else:
            _logger.warning("Unexisting map method")

    def _updateMotorAxis(self):
        """
        Update dimensions used for orientation distribution and mosaicity maps.
        This is only used on datasets with more than two dimensions.
        """

        self.dimensions = self._chooseDimensionWidget.dimension
        self.ori_dist, self.hsv_key = self.dataset.compute_mosaicity_colorkey(
            dimensions=self.dimensions, third_motor=0
        )
        self.mosaicity = self.compute_mosaicity()

        self._thirdMotorCB.clear()
        third_motor = self.dataset.dims.get(
            int(numpy.setdiff1d(range(3), self.dimensions)[0])
        )

        self._curvesColormap = Colormap(
            name="temperature",
            vmin=numpy.min(self.ori_dist),
            vmax=numpy.max(self.ori_dist),
        )
        state = self._thirdMotorCB.blockSignals(True)
        self._thirdMotorCB.addItems(numpy.array(third_motor.unique_values, dtype=str))
        self._thirdMotorCB.blockSignals(state)
        self._checkboxStateChanged()
        self._updatePlot(self._methodCB.currentText())

    def _updateThirdMotor(self, index=-1):
        """
        Update orientation distribution according to the third motor chosen
        """
        if index > -1:
            self.ori_dist, self.hsv_key = self.dataset.compute_mosaicity_colorkey(
                dimensions=self._chooseDimensionWidget.dimension, third_motor=index
            )

            self._curvesColormap = Colormap(
                name="temperature",
                vmin=numpy.min(self.ori_dist),
                vmax=numpy.max(self.ori_dist),
            )
            self._updatePlot(self._methodCB.currentText())

    def _opticolor(self, img, minc, maxc):
        img = img.copy()
        Cnn = img[~numpy.isnan(img)]
        sortC = sorted(Cnn)
        Imin = sortC[int(numpy.floor(len(sortC) * minc))]
        Imax = sortC[int(numpy.floor(len(sortC) * maxc))]
        img[img > Imax] = Imax
        img[img < Imin] = Imin

        return medfilt2d(img)

    def exportMaps(self):
        """
        Creates dictionay with maps information and exports it to a nexus file
        """
        if self.dataset.transformation:
            axes = [
                self.dataset.transformation.yregular,
                self.dataset.transformation.xregular,
            ]
            axes_names = ["y", "x"]
            axes_long_names = [
                self.dataset.transformation.label,
                self.dataset.transformation.label,
            ]
        else:
            axes = None
            axes_names = None
            axes_long_names = None

        if self.dataset and self.dataset.dims.ndim > 1:
            nx = {
                "entry": {
                    Method.MOSAICITY.value: create_nxdata_dict(
                        self.mosaicity,
                        Method.MOSAICITY.value,
                        axes,
                        axes_names,
                        axes_long_names,
                        rgba=True,
                    ),
                    Method.ORI_DIST.value: {
                        "key": {
                            "image": hsv_to_rgb(self.hsv_key),
                            "origin": self._contoursPlot.getImage().getOrigin(),
                            "scale": self._contoursPlot.getImage().getScale(),
                            "xlabel": self._contoursPlot.getImage().getXLabel(),
                            "ylabel": self._contoursPlot.getImage().getYLabel(),
                            "image@interpretation": "rgba-image",
                        },
                        "curves": self._curves,
                    },
                    "@NX_class": "NXentry",
                    "@default": Method.MOSAICITY.value,
                },
                "@NX_class": "NXroot",
                "@default": "entry",
            }

            for axis, dim in self.dataset.dims:
                nx["entry"][dim.name] = {"@NX_class": "NXcollection"}
                for i in range(4):
                    nx["entry"][dim.name][Method.values()[i]] = create_nxdata_dict(
                        self._moments[axis][i],
                        Method.values()[i],
                        axes,
                        axes_names,
                        axes_long_names,
                    )
        else:
            nx = {
                "entry": {"@NX_class": "NXentry"},
                "@NX_class": "NXroot",
                "@default": "entry",
            }

            for i in range(4):
                nx["entry"][Method.values()[i]] = create_nxdata_dict(
                    self._moments[0][i],
                    Method.values()[i],
                    axes,
                    axes_names,
                    axes_long_names,
                )
            nx["entry"]["@default"] = Method.COM.value

        fileDialog = qt.QFileDialog()

        fileDialog.setFileMode(fileDialog.AnyFile)
        fileDialog.setAcceptMode(fileDialog.AcceptSave)
        fileDialog.setOption(fileDialog.DontUseNativeDialog)
        fileDialog.setDefaultSuffix(".h5")
        if fileDialog.exec_():
            dicttonx(nx, fileDialog.selectedFiles()[0])

    def _addImage(self, plot, image):
        if self.dataset.transformation is None:
            plot.addImage(image, xlabel="pixels", ylabel="pixels")
            return
        if self.dataset.transformation.rotate:
            image = numpy.rot90(image, 3)
        plot.addImage(
            image,
            origin=self.dataset.transformation.origin,
            scale=self.dataset.transformation.scale,
            xlabel=self.dataset.transformation.label,
            ylabel=self.dataset.transformation.label,
        )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "07/12/2021"

import numpy
import logging

from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import StackView, Plot2D
from silx.utils.enum import Enum as _Enum
from silx.io.dictdump import dicttonx
from silx.image.marchingsquares import find_contours

import darfix
from darfix.core.mapping import fit_rocking_curve, fit_2d_rocking_curve
from darfix.core.dataset import Operation
from darfix.io.utils import create_nxdata_dict
from .operationThread import OperationThread

_logger = logging.getLogger(__file__)


class Maps(_Enum):
    """
    Different maps that can be showed after fitting the data
    """

    AMPLITUDE = "Amplitude"
    FWHM = "FWHM"
    PEAK = "Peak position"
    BACKGROUND = "Background"
    RESIDUALS = "Residuals"


class Maps_2D(_Enum):
    """
    Different maps that can be showed after fitting the data
    """

    AMPLITUDE = "Amplitude"
    PEAK_X = "Peak position first motor"
    PEAK_Y = "Peak position second motor"
    FWHM_X = "FWHM first motor"
    FWHM_Y = "FWHM second motor"
    BACKGROUND = "Background"
    CORRELATION = "Correlation"
    RESIDUALS = "Residuals"


class RockingCurvesWidget(qt.QMainWindow):
    """
    Widget to apply fit to a set of images and plot the amplitude, fwhm, peak position, background and residuals maps.
    """

    sigFitted = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        widget = qt.QWidget(parent=self)
        layout = qt.QGridLayout()

        self._sv = StackView(parent=self, position=True)
        self._sv.setColormap(Colormap(name=darfix.config.DEFAULT_COLORMAP_NAME))
        self._plot = Plot2D(parent=self)
        self._plot.setDefaultColormap(Colormap(name="cividis", normalization="linear"))
        self._plot.setGraphTitle("Rocking curves")
        self._plot.setGraphXLabel("Degrees")
        intLabel = qt.QLabel("Intensity threshold:")
        self._intThresh = "15"
        self._intThreshLE = qt.QLineEdit(self._intThresh)
        self._intThreshLE.setValidator(qt.QDoubleValidator())
        self._computeFit = qt.QPushButton("Fit data")
        self._computeFit.clicked.connect(self._grainPlot)
        self._abortFit = qt.QPushButton("Abort")
        self._abortFit.clicked.connect(self.__abort)
        spacer1 = qt.QWidget(parent=self)
        spacer1.setLayout(qt.QVBoxLayout())
        spacer1.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self._motorValuesCheckbox = qt.QCheckBox("Use motor values")
        self._motorValuesCheckbox.setChecked(True)
        self._motorValuesCheckbox.stateChanged.connect(self._checkboxStateChanged)
        self._centerDataCheckbox = qt.QCheckBox("Center angle values")
        self._centerDataCheckbox.setEnabled(False)
        self._centerDataCheckbox.stateChanged.connect(self._checkboxStateChanged)
        self._parametersLabel = qt.QLabel("")
        self._plotMaps = Plot2D(self)
        self._plotMaps.setDefaultColormap(
            Colormap(name="cividis", normalization="linear")
        )
        self._plotMaps.hide()
        self._methodCB = qt.QComboBox(self)
        self._methodCB.hide()
        self._exportButton = qt.QPushButton("Export maps")
        self._exportButton.hide()
        self._exportButton.clicked.connect(self.exportMaps)

        layout.addWidget(self._sv, 0, 0, 1, 2)
        layout.addWidget(self._plot, 0, 2, 1, 2)
        layout.addWidget(self._parametersLabel, 1, 2, 1, 2)
        layout.addWidget(self._motorValuesCheckbox, 2, 2, 1, 1)
        layout.addWidget(self._centerDataCheckbox, 2, 3, 1, 1)
        layout.addWidget(intLabel, 3, 0, 1, 1)
        layout.addWidget(self._intThreshLE, 3, 1, 1, 1)
        layout.addWidget(self._computeFit, 3, 2, 1, 2)
        layout.addWidget(self._abortFit, 3, 2, 1, 2)
        layout.addWidget(self._methodCB, 4, 0, 1, 4)
        layout.addWidget(self._plotMaps, 5, 0, 1, 4)
        layout.addWidget(self._exportButton, 6, 0, 1, 5)
        self._abortFit.hide()
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter.

        :param Dataset dataset: dataset
        """
        self._parent = parent
        self.dataset = dataset
        self.indices = indices
        self._update_dataset = dataset
        self._residuals = None
        self._thread = OperationThread(self, self.dataset.apply_fit)
        self.setStack()
        self._methodCB.clear()
        oldState = self._methodCB.blockSignals(True)
        if self.dataset.dims.ndim == 2:
            self._methodCB.addItems(Maps_2D.values())
            self._methodCB.currentTextChanged.connect(self._update2DPlot)
        else:
            self._methodCB.addItems(Maps.values())
            self._methodCB.currentTextChanged.connect(self._updatePlot)
        self._methodCB.blockSignals(oldState)
        self._sv.getPlotWidget().sigPlotSignal.connect(self._mouseSignal)
        self._sv.sigFrameChanged.connect(self._addPoint)

        if self.dataset.title != "":
            self._sv.setTitleCallback(lambda idx: self.dataset.title)

    def setStack(self, dataset=None):
        """
        Sets new data to the stack.
        Mantains the current frame showed in the view.

        :param Dataset dataset: if not None, data set to the stack will be from the given dataset.
        """
        if dataset is None:
            dataset = self.dataset
        nframe = self._sv.getFrameNumber()
        if self.indices is None:
            self._sv.setStack(dataset.get_data() if dataset is not None else None)
        else:
            self._sv.setStack(
                dataset.get_data(self.indices) if dataset is not None else None
            )
        self._sv.setFrameNumber(nframe)

    def getStackViewColormap(self):
        """
        Returns the colormap from the stackView

        :rtype: silx.gui.colors.Colormap
        """
        return self._sv.getColormap()

    def setStackViewColormap(self, colormap):
        """
        Sets the stackView colormap

        :param colormap: Colormap to set
        :type colormap: silx.gui.colors.Colormap
        """
        self._sv.setColormap(colormap)

    def _mouseSignal(self, info):
        """
        Method called when a signal from the stack is called
        """
        if info["event"] == "mouseClicked":
            # In case the user has clicked on a pixel in the stack
            data = self.dataset.get_data(self.indices)
            px = info["x"]
            py = info["y"]
            # Show vertical and horizontal lines for clicked pixel
            self._sv.getPlotWidget().addCurve(
                (px, px), (0, data.shape[1]), legend="x", color="r"
            )
            self._sv.getPlotWidget().addCurve(
                (0, data.shape[2]), (py, py), legend="y", color="r"
            )
            self.plotRockingCurves(px, py)

    def _addPoint(self, i=None):
        """
        Slot to add curve for frame number in rocking curves plot.

        :param int i: frame number
        """
        xc = self._sv.getPlotWidget().getCurve("x")
        if xc:
            px = xc.getXData()[0]
            py = self._sv.getPlotWidget().getCurve("y").getYData()[0]
            self.plotRockingCurves(px, py)

    def _computeContours(self, image, origin=None, scale=None):
        polygons = []
        levels = []
        for i in numpy.linspace(numpy.min(image), numpy.max(image), 10):
            polygons.append(find_contours(image, i))
            levels.append(i)
        # xdim = self.dataset.dims.get(1)
        # ydim = self.dataset.dims.get(0)
        for ipolygon, polygon in enumerate(polygons):
            # iso contours
            for icontour, contour in enumerate(polygon):
                if len(contour) == 0:
                    continue
                # isClosed = numpy.allclose(contour[0], contour[-1])
                x = contour[:, 1]
                y = contour[:, 0]
                if scale is not None:
                    x *= scale[0]
                    y *= scale[1]
                    x += origin[0] + scale[0] / 2
                    y += origin[1] + scale[1] / 2
                legend = "poly{}.{}".format(icontour, ipolygon)
                self._plot.addCurve(
                    x=x,
                    y=y,
                    linestyle="-",
                    linewidth=2.0,
                    legend=legend,
                    resetzoom=False,
                    color="w",
                )

    def plotRockingCurves(self, px, py):
        """
        Plot rocking curves of data and fitted data at pixel (px, py).

        :param Data data: stack of images to plot
        :param px: x pixel
        :param py: y pixel
        """
        # Get rocking curves from data
        self._plot.clear()
        try:
            data = self.dataset.get_data(self.indices)
            if self.dataset.in_memory:
                y = data[:, int(py), int(px)]
            else:
                y = numpy.array([image[int(py), int(px)] for image in data])
        except IndexError:
            _logger.warning("Index out of bounds")
            return
        if self.dataset.dims.ndim == 2:
            image = numpy.zeros(self.dataset.nframes)
            image[self.indices] = y
            xdim = self.dataset.dims.get(1)
            ydim = self.dataset.dims.get(0)
            self._plot.remove(kind="curve")
            frameNumber = self._sv.getFrameNumber()
            x = [
                self.dataset.get_metadata_values(kind=ydim.kind, key=ydim.name),
                self.dataset.get_metadata_values(kind=xdim.kind, key=xdim.name),
            ]
            dotx = int(frameNumber / ydim.size)
            doty = frameNumber % ydim.size
            xscale = xdim.range[2]
            yscale = ydim.range[2]
            if self._motorValuesCheckbox.isChecked():
                origin = [xdim.range[0], ydim.range[0]]
                dotx = xdim.unique_values[dotx]
                doty = ydim.unique_values[doty]
            else:
                origin = (0.0, 0.0)
                if self._centerDataCheckbox.isChecked():
                    dotx -= int(xdim.size / 2)
                    doty -= int(ydim.size / 2)
                    origin = (
                        -xscale * int(xdim.size / 2),
                        -yscale * int(ydim.size / 2),
                    )
                dotx *= xscale
                doty *= yscale
            try:
                y_gauss, pars = fit_2d_rocking_curve(
                    [image, None], values=x, shape=[ydim.size, xdim.size]
                )
                if numpy.array_equal(y_gauss, image):
                    raise RuntimeError
                y_gauss = numpy.reshape(y_gauss, (xdim.size, ydim.size)).T
                self._computeContours(y_gauss, origin, (xscale, yscale))
                self._parametersLabel.setText(
                    "PEAK_X:{:.3f} PEAK_Y:{:.3f} FWHM_X:{:.3f} FWHM_Y:{:.3f} AMP:{:.3f} CORR:{:.3f} BG:{:.3f}".format(
                        *pars
                    )
                )
            except (TypeError, RuntimeError):
                y_gauss = y
                _logger.warning("Cannot fit")

            y = numpy.reshape(image, (xdim.size, ydim.size)).T
            self._plot.addImage(
                y,
                xlabel=xdim.name,
                ylabel=ydim.name,
                origin=origin,
                scale=(xscale, yscale),
            )
            self._plot.addCurve([dotx], [doty], symbol="o", legend="dot_o", color="b")
            self.x, self.x_gauss, self.y, self.y_gauss = x, x, y, y_gauss
        else:
            if self.dataset.dims.ndim == 0:
                x = numpy.arange(data.shape[0])
            else:
                dim = self.dataset.dims.get(0)
                if self._motorValuesCheckbox.isChecked():
                    x = numpy.array(
                        self.dataset.get_metadata_values(
                            kind=dim.kind, key=dim.name, indices=self.indices
                        )
                    )
                else:
                    scale = dim.range[2]
                    x = numpy.arange(data.shape[0]) * scale
                    if self._centerDataCheckbox.isChecked():
                        x -= int(dim.size / 2)

            item = [numpy.array(y), None]
            if self._centerDataCheckbox.isChecked():
                middle = (float(x[-1]) - float(x[0])) / 2
                # x = numpy.linspace(-middle, middle, len(x))
                x -= float(x[0]) + middle
            # Show rocking curves and fitted curve into plot
            self._plot.clear()
            self._plot.addCurve(x, y, legend="data", color="b")
            i = self._sv.getFrameNumber()
            try:
                y_gauss, pars = fit_rocking_curve(item, values=x, num_points=1000)
                self._parametersLabel.setText(
                    "AMP:{:.3f} PEAK:{:.3f} FWHM:{:.3f} BG:{:.3f}".format(*pars)
                )
            except TypeError:
                y_gauss = y
                _logger.warning("Cannot fit")

            # Add curves (points) for stackview frame number
            self.x_gauss = numpy.linspace(x[0], x[-1], len(y_gauss))
            self._plot.addCurve(self.x_gauss, y_gauss, legend="fit", color="r")
            self._plot.addCurve([x[i]], [y[i]], symbol="o", legend="dot_o", color="b")
            i_gauss = i * int((len(y_gauss) - 1) / (len(x) - 1))
            self._plot.addCurve(
                [self.x_gauss[i_gauss]],
                [y_gauss[i_gauss]],
                symbol="o",
                legend="dot_fit",
                color="r",
            )
            self.x, self.y, self.y_gauss = x, y, y_gauss

    def _grainPlot(self):
        """
        Method called when button for computing fit is clicked
        """
        self._computeFit.hide()
        self._intThresh = self._intThreshLE.text()
        self.sigFitted.emit()
        self._thread.setArgs(self.indices, float(self._intThresh))
        self._thread.finished.connect(self._updateData)
        self._thread.start()
        self._abortFit.show()

    @property
    def residuals(self):
        if self._residuals is None:
            self._residuals = numpy.sqrt(
                numpy.subtract(
                    self._update_dataset.zsum(self.indices),
                    self.dataset.zsum(self.indices),
                )
                ** 2
            )
        return self._residuals

    def _updatePlot(self, method):
        """
        Updates the plots with the chosen method
        """
        try:
            method = Maps(method)
            title = self.dataset.title
            if title != "":
                title += " - "
            if method == Maps.AMPLITUDE:
                self._plotMaps.setGraphTitle(title + Maps.AMPLITUDE.value)
                self._addImage(self.maps[0])
            elif method == Maps.FWHM:
                self._plotMaps.setGraphTitle(title + Maps.FWHM.value)
                self._addImage(self.maps[2])
            elif method == Maps.PEAK:
                com = self.maps[1]
                com[numpy.isnan(com)] = min(com[~numpy.isnan(com)])
                self._plotMaps.setGraphTitle(title + Maps.PEAK.value)
                self._addImage(com)
            elif method == Maps.BACKGROUND:
                self._plotMaps.setGraphTitle(title + Maps.BACKGROUND.value)
                self._addImage(self.maps[3])
            elif method == Maps.RESIDUALS:
                self._plotMaps.setGraphTitle(title + Maps.RESIDUALS.value)
                self._addImage(self.residuals)
        except ValueError:
            _logger.warning("Unexisting map method")

    def _update2DPlot(self, method):
        """
        Updates the plots with the chosen method
        """
        try:
            method = Maps_2D(method)
            title = self.dataset.title
            if title != "":
                title += " - "
            if method == Maps_2D.PEAK_X:
                com = self.maps[0]
                com[numpy.isnan(com)] = min(com[~numpy.isnan(com)])
                self._plotMaps.setGraphTitle(title + Maps_2D.PEAK_X.value)
                self._addImage(com)
            elif method == Maps_2D.PEAK_Y:
                com = self.maps[1]
                com[numpy.isnan(com)] = min(com[~numpy.isnan(com)])
                self._plotMaps.setGraphTitle(title + Maps_2D.PEAK_Y.value)
                self._addImage(com)
            elif method == Maps_2D.FWHM_X:
                self._plotMaps.setGraphTitle(title + Maps_2D.FWHM_X.value)
                self._addImage(self.maps[2])
            elif method == Maps_2D.FWHM_Y:
                self._plotMaps.setGraphTitle(title + Maps_2D.FWHM_Y.value)
                self._addImage(self.maps[3])
            elif method == Maps_2D.AMPLITUDE:
                self._plotMaps.setGraphTitle(title + Maps_2D.AMPLITUDE.value)
                self._addImage(self.maps[4])
            elif method == Maps_2D.CORRELATION:
                self._plotMaps.setGraphTitle(title + Maps_2D.CORRELATION.value)
                self._addImage(self.maps[5])
            elif method == Maps_2D.BACKGROUND:
                self._plotMaps.setGraphTitle(title + Maps_2D.BACKGROUND.value)
                self._addImage(self.maps[6])
            elif method == Maps_2D.RESIDUALS:
                self._plotMaps.setGraphTitle(title + Maps_2D.RESIDUALS.value)
                self._addImage(self.residuals)
        except ValueError:
            _logger.warning("Unexisting map method")

    def __abort(self):
        self._abortFit.setEnabled(False)
        self.dataset.stop_operation(Operation.FIT)

    def _updateData(self):
        """
        Method called when fit computation has finished
        """
        self._thread.finished.disconnect(self._updateData)
        self._abortFit.hide()
        self._computeFit.show()
        if self._thread.data:
            self._update_dataset, self.maps = self._thread.data
            assert self._update_dataset is not None
            if self.dataset.dims.ndim == 2:
                self._update2DPlot(self._methodCB.currentText())
            else:
                self._updatePlot(self._methodCB.currentText())
            self._plotMaps.show()
            self._methodCB.show()
            self._exportButton.show()
        else:
            print("\nCorrection aborted")

    def _wholeStack(self):
        self.setStack(self.dataset)
        self._addPoint()

    def _checkboxStateChanged(self):
        """
        Update widgets linked to the checkbox state
        """
        self._centerDataCheckbox.setEnabled(not self._motorValuesCheckbox.isChecked())
        xc = self._sv.getPlotWidget().getCurve("x")
        if xc:
            px = xc.getXData()[0]
            py = self._sv.getPlotWidget().getCurve("y").getYData()[0]
            self.plotRockingCurves(px, py)
        if Maps(self._methodCB.currentText()) == Maps.PEAK:
            self._updatePlot(Maps.PEAK)

    @property
    def intThresh(self):
        return self._intThresh

    @intThresh.setter
    def intThresh(self, intThresh):
        self._intThresh = intThresh
        self._intThreshLE.setText(intThresh)

    def exportMaps(self):
        """
        Creates dictionay with maps information and exports it to a nexus file
        """
        entry = "entry"

        nx = {
            entry: {"@NX_class": "NXentry"},
            "@NX_class": "NXroot",
            "@default": "entry",
        }

        if self.dataset.transformation:
            axes = [
                self.dataset.transformation.yregular,
                self.dataset.transformation.xregular,
            ]
            axes_names = ["y", "x"]
            axes_long_names = [
                self.dataset.transformation.label,
                self.dataset.transformation.label,
            ]
        else:
            axes = None
            axes_names = None
            axes_long_names = None

        if self.dataset.dims.ndim == 2:
            for i, _map in enumerate(Maps_2D.values()):
                if _map == Maps_2D.RESIDUALS.value:
                    nx["entry"][_map] = create_nxdata_dict(
                        self.residuals, _map, axes, axes_names, axes_long_names
                    )
                else:
                    nx["entry"][_map] = create_nxdata_dict(
                        self.maps[i], _map, axes, axes_names, axes_long_names
                    )
            nx["entry"]["@default"] = Maps_2D.AMPLITUDE.value
        else:
            for i, _map in enumerate(Maps.values()):
                if _map == Maps.RESIDUALS.value:
                    nx["entry"][_map] = create_nxdata_dict(
                        self.residuals, _map, axes, axes_names, axes_long_names
                    )
                else:
                    nx["entry"][_map] = create_nxdata_dict(
                        self.maps[i], _map, axes, axes_names, axes_long_names
                    )
            nx["entry"]["@default"] = Maps.AMPLITUDE.value

        fileDialog = qt.QFileDialog()

        fileDialog.setFileMode(fileDialog.AnyFile)
        fileDialog.setAcceptMode(fileDialog.AcceptSave)
        fileDialog.setOption(fileDialog.DontUseNativeDialog)
        fileDialog.setDefaultSuffix(".h5")
        if fileDialog.exec_():
            dicttonx(nx, fileDialog.selectedFiles()[0])

    def _addImage(self, image):
        if self.dataset.transformation is None:
            self._plotMaps.addImage(image, xlabel="pixels", ylabel="pixels")
            return
        if self.dataset.transformation.rotate:
            image = numpy.rot90(image, 3)
        self._plotMaps.addImage(
            image,
            origin=self.dataset.transformation.origin,
            scale=self.dataset.transformation.scale,
            xlabel=self.dataset.transformation.label,
            ylabel=self.dataset.transformation.label,
        )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "22/12/2020"

import numpy

from silx.gui import qt
from silx.utils.enum import Enum as _Enum

from .operationThread import OperationThread
from .displayComponentsWidget import DisplayComponentsWidget


class Method(_Enum):
    """
    Different blind source separation approaches that can be applied
    """

    PCA = (
        "The process of computing the principal components \n"
        "and using them to perform a change of basis on the data"
    )
    NICA = "Find components independent from each other and non-negative"
    NMF = (
        "Non-negative matrix factorization factorizes the data matrix into \n"
        "two matrices, with the property that all three matrices have no negative elements"
    )
    NICA_NMF = "Apply Non-negative ICA followed by NMF"


class BSSWidget(qt.QMainWindow):
    """
    Widget to apply blind source separation.
    """

    sigComputed = qt.Signal(Method, int)

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # Method widget
        methodLabel = qt.QLabel("Method: ")
        self.methodCB = qt.QComboBox()
        for member in Method.members():
            self.methodCB.addItem(member.name)
            self.methodCB.setItemData(
                self.methodCB.count() - 1, member.value, qt.Qt.ToolTipRole
            )
        # Number of components
        nComponentsLabel = qt.QLabel("Num comp:")
        self.nComponentsLE = qt.QLineEdit("1")
        self.nComponentsLE.setValidator(qt.QIntValidator())
        # Compute BSS with the number of components
        self.computeButton = qt.QPushButton("Compute")
        self.computeButton.setEnabled(False)
        self.computeButton.clicked.connect(self._computeBSS)
        # Detect optimal number of components
        self.detectButton = qt.QPushButton("Detect number of components")
        self.detectButton.setEnabled(False)
        self.detectButton.clicked.connect(self._detectComp)

        # Add widgets to layout
        layout = qt.QGridLayout()
        layout.addWidget(methodLabel, 0, 0, 1, 1)
        layout.addWidget(self.methodCB, 0, 1, 1, 1)
        layout.addWidget(nComponentsLabel, 0, 2, 1, 1)
        layout.addWidget(self.nComponentsLE, 0, 3, 1, 1)
        layout.addWidget(self.computeButton, 0, 4, 1, 1)
        layout.addWidget(self.detectButton, 1, 4, 1, 1)
        # Top Widget with the type of bss and the number of components to compute
        top_widget = qt.QWidget(self)
        top_widget.setLayout(layout)

        # Widget to display the components
        self._displayComponentsWidget = DisplayComponentsWidget()
        self._displayComponentsWidget.hide()

        # Main widget is a Splitter with the top widget and the displayComponentsWidget
        self.splitter = qt.QSplitter(qt.Qt.Vertical)
        self.splitter.addWidget(top_widget)
        self.splitter.addWidget(self._displayComponentsWidget)
        self.setCentralWidget(self.splitter)

    def hideButton(self):
        self._computeB.hide()

    def showButton(self):
        self._computeB.show()

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter.

        :param Dataset dataset: dataset
        """
        self._parent = parent
        self.dataset = dataset
        self.indices = indices
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset
        self.computeButton.setEnabled(True)
        self.detectButton.setEnabled(True)

    def _updateDataset(self, widget, dataset):
        self._parent._updateDataset(widget, dataset)
        self.dataset = dataset

    def _computeBSS(self):
        """
        Computes blind source separation with the chosen method.
        """
        self.computeButton.setEnabled(False)
        self.detectButton.setEnabled(False)
        self.nComponentsLE.setEnabled(False)
        method = Method[self.methodCB.currentText()]
        n_comp = int(self.nComponentsLE.text())
        self.sigComputed.emit(method, n_comp)
        if method == Method.PCA:
            self._thread = OperationThread(self, self.dataset.pca)
            self._thread.setArgs(n_comp, indices=self.indices)
        elif method == Method.NICA:
            self._thread = OperationThread(self, self.dataset.nica)
            self._thread.setArgs(n_comp, indices=self.indices, error_step=5)
        elif method == Method.NMF:
            self._thread = OperationThread(self, self.dataset.nmf)
            self._thread.setArgs(n_comp, indices=self.indices, error_step=5)
        elif method == Method.NICA_NMF:
            self._thread = OperationThread(self, self.dataset.nica_nmf)
            self._thread.setArgs(n_comp, indices=self.indices, error_step=5)
        else:
            raise ValueError("BSS method not managed")

        self._thread.finished.connect(self._displayComponents)
        self._thread.start()

    def _displayComponents(self):
        self._thread.finished.disconnect(self._displayComponents)
        comp, self.W = self._thread.data
        n_comp = int(self.nComponentsLE.text())
        if comp.shape[0] < n_comp:
            n_comp = comp.shape[0]
            msg = qt.QMessageBox()
            msg.setIcon(qt.QMessageBox.Information)
            msg.setText("Found only {0} components".format(n_comp))
            msg.setStandardButtons(qt.QMessageBox.Ok)
            msg.exec_()
        shape = self.dataset.get_data()[0].shape
        self.comp = comp.reshape(n_comp, shape[0], shape[1])
        self._displayComponentsWidget.show()
        self.computeButton.setEnabled(True)
        self.nComponentsLE.setEnabled(True)
        self.detectButton.setEnabled(True)
        if self.bg_indices is not None:
            # If filter data is activated, the matrix W has reduced dimensionality, so reshaping is not possible
            # Create empty array with shape the total number of frames
            W = numpy.zeros((self.dataset.nframes, n_comp))
            # Set actual values of W where threshold of filter is True
            W[self.indices] = self.W
            self.W = W
        self._displayComponentsWidget.setComponents(
            self.comp,
            self.W,
            self.dataset.dims,
            self.dataset.get_dimensions_values(),
            self.dataset.title,
        )

    def _detectComp(self):
        self.detectButton.setEnabled(False)
        self.computeButton.setEnabled(False)
        self._thread = OperationThread(self, self.dataset.pca)
        self._thread.setArgs(return_vals=True)
        self._thread.finished.connect(self._setNumComp)
        self._thread.start()

    def _setNumComp(self):
        self._thread.finished.disconnect(self._setNumComp)
        vals = self._thread.data
        vals /= numpy.sum(vals)
        components = len(vals[vals > 0.01])
        self.detectButton.setEnabled(True)
        self.computeButton.setEnabled(True)
        self.nComponentsLE.setText(str(components))

    def getDisplayComponentsWidget(self):
        return self._displayComponentsWidget



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "09/12/2021"


from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow

import darfix
from darfix.gui.utils import ChooseDimensionWidget
from .operationThread import OperationThread


class ProjectionWidget(qt.QMainWindow):
    """
    Widget to apply a projection to the chosen dimension.
    """

    sigComputed = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self._dimension = None

        self._sv = StackViewMainWindow()
        self._sv.setColormap(
            Colormap(name=darfix.config.DEFAULT_COLORMAP_NAME, normalization="linear")
        )
        self._chooseDimensionWidget = ChooseDimensionWidget(
            self, vertical=False, values=False, _filter=False
        )
        self._projectButton = qt.QPushButton("Project data")
        self._projectButton.setEnabled(False)
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.setEnabled(False)
        layout = qt.QGridLayout()
        layout.addWidget(self._chooseDimensionWidget, 0, 0, 1, 2)
        layout.addWidget(self._projectButton, 1, 1)
        layout.addWidget(self._sv, 2, 0, 1, 2)
        layout.addWidget(self._buttons, 3, 1)
        self._sv.hide()
        widget = qt.QWidget()
        widget.setLayout(layout)

        self._buttons.accepted.connect(self.sigComputed.emit)
        self._projectButton.clicked.connect(self._projectData)

        self.setCentralWidget(widget)

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter.

        :param Dataset dataset: dataset
        """
        if dataset is not None:
            self._parent = parent
            self._dataset = dataset
            self._update_dataset = dataset
            self.indices = indices
            self.bg_indices = bg_indices
            self.bg_dataset = bg_dataset
            self._projectButton.setEnabled(True)
            if self._dataset.dims.ndim > 1:
                self._buttons.setEnabled(True)

            self._chooseDimensionWidget.setDimensions(self._dataset.dims)
            self._chooseDimensionWidget._updateState(True)
            for i in range(1, self._dataset.dims.ndim - 1):
                self._chooseDimensionWidget.dimensionWidgets[i][0].addItem("None")
            self._thread = OperationThread(self, self._dataset.project_data)
            if self._dataset.title != "":
                self._sv.setTitleCallback(lambda idx: self._dataset.title)

    def getDataset(self, update=None):
        return self._update_dataset, self.indices, self.bg_indices, self.bg_dataset

    def clearStack(self):
        self._sv.setStack(None)
        self._projectButton.setEnabled(False)

    @property
    def dimension(self):
        return self._dimension

    @dimension.setter
    def dimension(self, dimension):
        self._dimension = dimension

    def _projectData(self):
        self._projectButton.setEnabled(False)
        self.dimension = [self._chooseDimensionWidget.dimension[0]]
        for i in range(1, len(self._chooseDimensionWidget.dimension)):
            self.dimension += [self._chooseDimensionWidget.dimension[i]]
        self._thread.setArgs(dimension=self.dimension, indices=self.indices)
        self._thread.finished.connect(self._updateData)
        self._thread.start()

    def _updateDataset(self, widget, dataset):
        self._parent._updateDataset(widget, dataset)
        self._dataset = dataset

    def _updateData(self):
        self._projectButton.setEnabled(True)
        self._thread.finished.disconnect(self._updateData)
        if self._thread.data:
            del self._update_dataset
            self._update_dataset = self._thread.data
            self.indices, self.bg_indices = None, None
            self._sv.show()
            self._sv.setStack(self._update_dataset.get_data())



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "28/04/2020"

import numpy
import unittest

from darfix.decomposition.test.utils import images, sampler
from darfix.decomposition.nmf import NMF


class TestNMF(unittest.TestCase):
    """Tests for `nmf.py`."""

    def setUp(self):
        self.rstate = numpy.random.RandomState(seed=1000)
        self.images = self.rstate.random((100, 1000))

    def test_fit_transform(self):
        resources = ["circle", "star", "pentagon", "square"]
        num_images = 100
        means = [15, 30, 45, 60]
        tol = 0.5
        sample = sampler(resources, means)

        X = numpy.array([sample(i).flatten() for i in range(num_images)])

        nmf = NMF(X, 4)
        H = self.rstate.random((nmf.num_components, nmf.num_features)) + 10**-4
        W = self.rstate.random((nmf.num_samples, nmf.num_components)) + 10**-4
        nmf.fit_transform(max_iter=500, H=H, W=W)

        stack = numpy.asarray(images(resources))
        for img in stack:
            img = img / numpy.linalg.norm(img)
            found = False
            for component in nmf.H:
                comp = component.reshape(img.shape)
                comp = comp / numpy.linalg.norm(comp)
                err = numpy.linalg.norm(comp - img)
                if err < tol:
                    found = True
                    break
            self.assertTrue(found)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "14/05/2020"


import numpy
import scipy.sparse

from darfix.io import utils


class Base:
    """
    Base class for decomposition package.

    :param data: Numpy array or Hdf5 dataset with images in the rows and
        pixels in the columns.
    :param num_components: Number of components to keep, defaults to None
    :type num_components: Union[None,int], optional
    :param indices: The indices of the values to use, defaults to None
    :type indices: Union[None,array_like], optional
    :param epsilon: Convergence tolerance, defaults to 1e-07
    :type epsilon: float
    """

    def __init__(self, data, num_components=None, indices=None, epsilon=1e-7):
        self._data = data
        self._epsilon = epsilon

        self._num_samples, self._num_features = self.data.shape
        if indices is not None:
            self._num_samples = len(indices)
        self._indices = (
            sorted(indices) if indices is not None else numpy.arange(len(data))
        )
        self._num_components = (
            num_components
            if num_components
            else min(self.num_samples, self.num_features)
        )

    @property
    def data(self):
        return self._data

    @property
    def indices(self):
        return self._indices

    @property
    def num_samples(self):
        return self._num_samples

    @property
    def num_features(self):
        return self._num_features

    @property
    def num_components(self):
        return self._num_components

    def _init_w(self):

        self.W = numpy.random.random((self.num_samples, self.num_components))

        """ Add small value for faster convergence with non-zero values"""
        self.W += 10**-4

    def _init_h(self):

        self.H = numpy.random.random((self.num_components, self.num_features))

        """ Add small value for faster convergence with non-zero values"""
        self.H += 10**-4

    def _update_h(self):

        pass

    def _update_w(self):

        pass

    def frobenius_norm(self, chunks=200):
        """Frobenius norm (||data - WH||) of a data matrix and a low rank
        approximation given by WH. Minimizing the Fnorm is the most common
        optimization criterion for matrix factorization methods.
        Returns:
        -------
        frobenius norm: F = ||data - WH||
        """
        # check if W and H exist
        if hasattr(self, "H") and hasattr(self, "W"):
            err = 0
            for i in range(0, self.data.shape[1], chunks):
                if scipy.sparse.issparse(self.data):
                    tmp = self.data[self.indices, i : i + chunks] - (
                        self.W * self.H[:, i : i + chunks]
                    )
                    err = tmp.multiply(tmp).sum()
                else:
                    err += numpy.sum(
                        (
                            self.data[self.indices, i : i + chunks]
                            - numpy.dot(self.W, self.H[:, i : i + chunks])
                        )
                        ** 2
                    )
            err = numpy.sqrt(err)
        else:
            err = None

        return err

    def squared_frobenius_norm(self, chunks=200):
        """Frobenius norm (||data - WH||) of a data matrix and a low rank
        approximation given by WH. Minimizing the Fnorm is the most common
        optimization criterion for matrix factorization methods.
        Returns:
        -------
        frobenius norm: F = ||data - WH||
        """
        # check if W and H exist
        if hasattr(self, "H") and hasattr(self, "W"):
            err = 0
            for i in range(0, self.data.shape[1], chunks):
                if scipy.sparse.issparse(self.data):
                    tmp = self.data[self.indices, i : i + chunks] - (
                        self.W * self.H[:, i : i + chunks]
                    )
                    err = tmp.multiply(tmp).sum()
                else:
                    err += numpy.sum(
                        (
                            self.data[self.indices, i : i + chunks]
                            - numpy.dot(self.W, self.H[:, i : i + chunks])
                        )
                        ** 2
                    )
        else:
            err = None
        return err

    def _converged(self, i):
        """
        If the optimization of the approximation is below the machine precision,
        return True.
        :param int i: index of the update step

        :return: if converged
        :rtype: bool
        """
        if self.ferr[i] and self.ferr[i - 1]:
            derr = numpy.abs(self.ferr[i] - self.ferr[i - 1]) / self._num_samples
            if derr < self._epsilon:
                return True
        return False

    def fit_transform(
        self,
        max_iter=100,
        error_step=None,
        compute_w=True,
        compute_h=True,
        norm="frobenius",
    ):
        """
        Fit to data, then transform it

        :param int max_iter: Maximum number of iterations, defaults to 100
        :type max_iter: int, optional
        :param error_step: If None, error is not computed, defaults to None
            Else compute error for every `error_step` iterations.
        :type error_step: Union[None,int], optional
        :param compute_w: When False, W is not computed, defaults to True
        :type compute_w: bool, optional
        :param compute_h: When False, H is not computed, defaults to True
        :type compute_h: bool, optional
        """

        self.ferr = []

        if compute_w:
            self._init_w()
        if compute_h:
            self._init_h()

        if max_iter > 1:
            utils.advancement_display(0, max_iter, "Updating decomposition matrices")

        for i in range(max_iter):
            if compute_w:
                self._update_w()
            if compute_h:
                self._update_h()

            if error_step:
                if not i % error_step:
                    if norm == "frobenius":
                        self.ferr.append(self.frobenius_norm())
                    elif norm == "squared_frobenius":
                        self.ferr.append(self.squared_frobenius_norm())

                    if i > error_step and self._converged(int(i / error_step)):
                        self.ferr = self.ferr[:i]
                        break
            if max_iter > 1:
                utils.advancement_display(
                    i + 1, max_iter, "Updating decomposition matrices"
                )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno", "J. Garriga"]
__license__ = "MIT"
__date__ = "24/09/2019"


from silx.gui import qt
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Input, Output
from darfix.core import utils
from darfix.core.dataset import Dimension
from darfix.gui.dimensionsWidget import DimensionWidget
from functools import partial
import logging
from darfix.core.process import DimensionDefinition

_logger = logging.getLogger(__file__)


class DimensionWidgetOW(OWWidget):
    """
    Widget used to define the calibration of the experimentation (select motor
    positions...)
    """

    name = "dimension definition"
    id = "orange.widgets.darfix.dimensiondefinition"
    description = "Define the dimension followed during the acquisition"
    icon = "icons/param_dims.svg"
    ewokstaskclass = DimensionDefinition

    priority = 4
    keywords = ["dataset", "calibration", "motor", "angle", "geometry"]

    # Inputs
    class Inputs:
        dataset = Input("dataset", tuple)

    # Outputs
    class Outputs:
        dataset = Output("dataset", tuple)

    want_main_area = False

    _dims = Setting(dict(), schema_only=True)

    def __init__(self):
        super().__init__()
        self._widget = DimensionWidget(parent=self)
        self.controlArea.layout().addWidget(self._widget)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self.buttons = qt.QDialogButtonBox(parent=self)
        self.buttons.setStandardButtons(types)
        self.controlArea.layout().addWidget(self.buttons)

        self.buttons.accepted.connect(self.validate)
        self.buttons.button(qt.QDialogButtonBox.Ok).setEnabled(False)

        # connect Signal/SLOT
        _callbackValid = partial(
            self.buttons.button(qt.QDialogButtonBox.Ok).setEnabled, True
        )
        self._widget.fitSucceed.connect(_callbackValid)
        _callbackInvalid = partial(
            self.buttons.button(qt.QDialogButtonBox.Ok).setDisabled, True
        )
        self._widget.fitFailed.connect(_callbackInvalid)

        # expose API
        self.setDims = self._widget.setDims

    @property
    def _ndim(self):
        return self._widget.ndim

    @property
    def _dataset(self):
        return self._widget.dataset

    @Inputs.dataset
    def setDataset(self, _input):
        """
        Input signal to set the dataset.
        """
        if _input is not None:
            dataset, update = _input
            self.buttons.button(qt.QDialogButtonBox.Ok).setEnabled(False)
            try:
                self._widget.setDataset(*dataset)
                # load properties
                _dims = utils.convertDictToDim(self._dims)
                self._widget.setDims(_dims)
            except ValueError as e:
                qt.QMessageBox.warning(
                    self, "Fail to setup dimension definition", str(e)
                )
            else:
                if update is None:
                    self.open()
                elif update != self:
                    self.Outputs.dataset.send(
                        ((self,) + self._widget.getDataset(), update)
                    )

    def _updateDataset(self, widget, dataset):
        self._widget._updateDataset(widget, dataset)

    def validate(self):
        """
        Tries to fit the dimensions into the dataset.
        """
        self.Outputs.dataset.send(((self,) + self._widget.getDataset(), None))
        self.updateProperties()
        OWWidget.accept(self)

    def updateProperties(self):
        """
        Save dimensions to Settings.
        """
        self._dims = {}
        if self._widget.ndim == 0:
            return
        else:
            for _dim in self._widget.dims.values():
                assert isinstance(_dim, Dimension)
                axis = _dim.axis
                self._dims[axis] = _dim.to_dict()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "22/12/2020"

import numpy

from silx.gui import qt
from silx.gui.plot import Plot1D

from .operationThread import OperationThread


class PCAWidget(qt.QMainWindow):
    """
    Widget to apply PCA to a set of images and plot the eigenvalues found.
    """

    sigComputed = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self._plot = Plot1D()
        self._plot.setDataMargins(0.05, 0.05, 0.05, 0.05)
        self._plot.setGraphXLabel("Components")
        self._plot.setGraphYLabel("Singular values")
        self.setCentralWidget(self._plot)

    def _computePCA(self):
        try:
            self._thread = OperationThread(self, self._dataset.pca)
            self._thread.setArgs(return_vals=True, indices=self.indices)
            self._thread.finished.connect(self._updateData)
            self._thread.start()
        except Exception as e:
            raise e

    def setDataset(
        self, parent, dataset, indices=None, bg_indices=None, bg_dataset=None
    ):
        """
        Dataset setter.

        :param Dataset dataset: dataset
        """
        self._parent = parent
        self._dataset = dataset
        self.indices = indices
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset
        self._plot.setGraphTitle(
            "Components representation of the dataset " + self._dataset.title
        )
        self._computePCA()

    def _updateDataset(self, widget, dataset):
        self._parent._updateDataset(widget, dataset)
        self._dataset = dataset

    def _updateData(self):
        """
        Plots the eigenvalues.
        """
        self._thread.finished.disconnect(self._updateData)
        vals = self._thread.data
        self._plot.show()
        self._plot.addCurve(numpy.arange(len(vals)), vals, symbol=".", linestyle=" ")
        self.sigComputed.emit()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "22/04/2020"

import numpy
import unittest

from scipy.stats import special_ortho_group

from darfix.decomposition.ipca import IPCA


class TestIPCA(unittest.TestCase):
    """Tests for `ipca.py`."""

    @classmethod
    def setUpClass(cls):
        cls.rstate = numpy.random.RandomState(seed=1000)

    def test_singular_values(self):
        data = self.rstate.random((100, 1000))
        ipca = IPCA(data, 50, num_components=3)
        self.assertEqual(ipca.singular_values, None)
        ipca.fit_transform()
        self.assertEqual(len(ipca.singular_values), 3)

    def test_fit_transform(self):
        # Number of features (dimension of the sample space)
        n = 10
        # Number of samples
        k = 100000
        # Number of principal components (dimension of the affine subspace)
        r = 5

        # Build principal components
        all_dims = numpy.arange(n)
        self.rstate.shuffle(all_dims)
        sub_dims = all_dims[:r]
        V = numpy.eye(n)[:, sub_dims]  # Subset of the standard basis vectors
        R = special_ortho_group.rvs(n)  # Random rotation in R^n
        V = numpy.dot(R, V)  # Principal components
        mean = self.rstate.rand(n)  # Affine space translation

        # Build observations
        D = numpy.diag(range(r, 0, -1)) ** 2  # Matrix of decreasing eigenvalues
        Z = self.rstate.normal(size=(k, r))  # Dimensionality-reduced samples
        Z = numpy.dot(Z, D)
        X = numpy.dot(Z, V.T) + mean  # Observations

        # Apply PCA to X to get Z, V
        ipca = IPCA(X, chunksize=10000, num_components=r)
        ipca.fit_transform()
        estimated_Z = numpy.round(ipca.W, 1)
        estimated_V = numpy.round(ipca.H.T, 1)

        numpy.testing.assert_allclose(
            numpy.abs(estimated_Z), numpy.abs(Z), rtol=0, atol=1
        )
        numpy.testing.assert_allclose(
            numpy.abs(estimated_V), numpy.abs(V), rtol=0, atol=1
        )

    def test_fit_transform_indices(self):
        # Number of features (dimension of the sample space)
        n = 10
        # Number of samples
        k = 100000
        # Indices
        indices = numpy.arange(50000)
        # Number of principal components (dimension of the affine subspace)
        r = 5

        # Build principal components
        all_dims = numpy.arange(n)
        self.rstate.shuffle(all_dims)
        sub_dims = all_dims[:r]
        V = numpy.eye(n)[:, sub_dims]  # Subset of the standard basis vectors
        R = special_ortho_group.rvs(n)  # Random rotation in R^n
        V = numpy.dot(R, V)  # Principal components
        mean = self.rstate.rand(n)  # Affine space translation

        # Build observations
        D = numpy.diag(range(r, 0, -1)) ** 2  # Matrix of decreasing eigenvalues
        Z = self.rstate.normal(size=(k, r))  # Dimensionality-reduced samples
        Z = numpy.dot(Z, D)
        X = numpy.dot(Z, V.T) + mean  # Observations

        # Apply PCA to X to get Z, V
        ipca = IPCA(X, chunksize=10000, num_components=r, indices=indices)
        ipca.fit_transform()
        estimated_Z = numpy.round(ipca.W, 1)
        estimated_V = numpy.round(ipca.H.T, 1)

        numpy.testing.assert_allclose(
            numpy.abs(estimated_Z), numpy.abs(Z[indices]), rtol=0, atol=1
        )
        numpy.testing.assert_allclose(
            numpy.abs(estimated_V), numpy.abs(V), rtol=0, atol=1
        )
